'''
   Copyright 2020 Matthew Ralston

   Licensed under the Apache License, Version 2.0 (the "License");
   you may not use this file except in compliance with the License.
   You may obtain a copy of the License at

       http://www.apache.org/licenses/LICENSE-2.0

   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   See the License for the specific language governing permissions and
   limitations under the License.

'''


import io
import sys
import os
import gzip
import tempfile
import yaml, json
from collections import deque, OrderedDict



def represent_ordereddict(dumper, data):
    """
    Thanks to Blender and EquipDev on StackOverflow for this handy method to pass to yaml.add_representer
    https://stackoverflow.com/a/16782282
    """
    value = []

    for item_key, item_value in data.items():
        node_key = dumper.represent_data(item_key)
        node_value = dumper.represent_data(item_value)

        value.append((node_key, node_value))

    return yaml.nodes.MappingNode(u'tag:yaml.org,2002:map', value)


def merge_metadata_lists(k, metadata_across_all_files, new_kmer_metadata):
    """
    Merge two 4**k metadata lists
    """
    if 4**k != len(new_kmer_metadata):
        raise TypeError("kmerdb.util.merge_metadata_lists() expects a new_kmer_metadata list to have 4^{0} elements. Got {1}".format(k, len(new_kmer_metadata)))

    for i, metadata in enumerate(new_kmer_metadata):
        metadata_across_all_files[i] = metadata_across_all_files[i] + metadata

    return metadata_across_all_files



'''
   Copyright 2020 Matthew Ralston

   Licensed under the Apache License, Version 2.0 (the "License");
   you may not use this file except in compliance with the License.
   You may obtain a copy of the License at

       http://www.apache.org/licenses/LICENSE-2.0

   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   See the License for the specific language governing permissions and
   limitations under the License.

'''



import logging
import argparse
import os
import sys
import yaml
import json
import time
from multiprocessing import cpu_count
from collections import OrderedDict


from Bio import bgzf

#import concurrent.futures

global logger
logger = None


def print_argv():
    argv = sys.argv
    sys.stderr.write(" ".join(argv[0:4]) + " ...\n")

def citation(arguments):
    import pkg_resources
    citation = None
    if pkg_resources.resource_exists('kmerdb', 'CITATION'):
        citation_fname = pkg_resources.resource_filename('kmerdb', 'CITATION')
        with open(citation_fname, 'w') as citation_file:
            citation_file.write("")


    sys.stderr.write("On the real, we gotta eat.")
    sys.stderr.write("Consider a +1 to keep it real...")

def index_file(arguments):
    from kmerdb import fileutil, index
    from kmerdb.config import DONE
    
    with fileutil.open(arguments.kdb, mode='r') as kdb:
        k = kdb.metadata['k']
    line_index = index.open(arguments.kdb, mode="xt", k=k)
    sys.stderr.write(DONE)


def shuf(arguments):
    import random
    
    from kmerdb import fileutil, index
    from kmerdb.config import DONE

    if os.path.splitext(arguments.kdb_in)[-1] != ".kdb":
        raise ValueError("kdb shuf requires an indexed .kdb file. Received '{0}'".format(arguments.kdb))
    else:
        arguments.kdb_index = arguments.kdb_in + "i"
    
        with fileutil.open(arguments.kdb_in, mode='r') as kdb:
            with index.open(arguments.kdb_index, mode='r') as kdbi:
                shuffled = list(range(len(kdbi.index)))
                random.shuffle(shuffled)
                random.shuffle(shuffled)
                with fileutil.open(arguments.kdb_out, mode='w', metadata=kdb.metadata) as kdb_out:
                    x = 1
                    for i in shuffled:
                        kmer_id, count, kmer_metadata = index.read_line(kdb, kdbi, i)
                        kdb_out.write("{0}\t{1}\t{2}".format(x, count, kmer_metadata))
                        x += 1
    sys.stderr.write(DONE)
    
def markov_probability(arguments):
    import pandas as pd
    import numpy as np
    from kmerdb import fileutil, index, probability, seqparser
    from kmerdb.config import DONE

    if os.path.splitext(arguments.kdb)[-1] != ".kdb":
        raise IOError("Model .kdb filepath does not end in '.kdb'")


    if index.has_index(arguments.kdb):
        arguments.kdbi = arguments.kdb + "i"
        #df = pd.DataFrame([], columns=["SequenceID", "Log_Odds_ratio", "p_of_seq"])
        profiles = np.array([], dtype="int64")
        with fileutil.open(arguments.kdb, 'r') as kdb:
            k = kdb.metadata['k']
            with index.open(arguments.kdbi, 'r') as kdbi:
                with seqparser.SeqParser(arguments.seqfile, arguments.fastq_block_size, k) as seqprsr:
                    recs = [r for r in seqprsr]
                    if seqprsr.fastq:
                        logger.debug("Read exactly b=={0}=={1} records from the {2} seqparser object".format(b, len(recs), s))
                        assert len(recs) == b, "The seqparser should return exactly {0} records at a time".format(b)
                    else:
                        logger.debug("Read {0} sequences from the {1} seqparser object".format(len(recs), seqprsr))
                        logger.debug("Skipping the block size assertion for fasta files")

                    while len(recs): # While the seqprsr continues to produce blocks of reads
                        # Do something here
                    
                        markov_probs = list(map(lambda p: [p["seq"].name, p["log_odds_ratio"], p["p_of_seq"]], [probability.markov_probability(seq, kdb, kdbi) for seq in recs]))

                        sys.stderr.write(json.dumps(markov_probs))
                        if profiles.shape == (0,):
                            profiles = np.array(markov_probs)
                        else:
                            np.append(profiles, markov_probs, axis=0)

                        recs = [r for r in seqprsr] # Essentially accomplishes an iteration in the file, wrapped by the seqparser.SeqParser class
        df = pd.DataFrame(profiles, columns=["SequenceID", "Log_Odds_ratio", "p_of_seq"])
        df.to_csv(sys.stdout, sep=arguments.delimiter, index=False)

    else:
        raise IndexError(".kdb file '{0}' has no corresponding index file. Please run 'kdb index -h' for details on index generation".format(arguments.kdb))

    
    sys.stderr.write(DONE)
                    
def distances(arguments):
    import pandas as pd
    import numpy as np
    from scipy.spatial.distance import pdist, squareform


    from kmerdb import fileutil, distance, config
    n = len(arguments.input)
    try:
        np.dtype(arguments.dtype)
    except TypeError as e:
        logger.error(e)
        logger.error("kmerdb encountered a TypeError, halting.")
        raise argparse.ArgumentError("Dtype '{0}' is invalid".format(arguments.dtype))

    if len(arguments.input) > 1:
        files = list(map(lambda f: fileutil.open(f, 'r'), arguments.input))
        logger.debug("Files: {0}".format(files))
        if not all(os.path.splitext(kdb)[-1] == ".kdb" for kdb in arguments.input):
            raise IOError("One or more parseable .kdb filepaths did not end in '.kdb'")
        dtypes = [x.profile.dtype.name for x in files]
        suggested_dtype = dtypes[0]
        
        ks = [kdbrdr.k for kdbrdr in files]
        suggested_k = ks[0]
        if not all(kdbrdr.k == suggested_k for kdbrdr in files):
            logger.error("Files: {0}".format(files))
            logger.error("Choices of k: {0}".format(ks))
            logger.error("By default the inferred value of k is used when k is not specified at the command line, which was {0}".format(suggested_k))
            raise TypeError("One or more files did not have k set to be equal to {0}".format(arguments.k))
        logger.debug("Files: {0}".format(files))
        
        if not all(kdbrdr.dtype == suggested_dtype for kdbrdr in files):
            raise TypeError("One or more files did not have dtype = {0}".format(suggested_dtype))
        data = [kdbrdr.profile for kdbrdr in files]
        #profiles = np.transpose(np.array(data, dtype=suggested_dtype))
        #profiles = np.array(data, dtype=suggested_dtype)
        profiles = np.transpose(data)
        logger.debug("Dammit, wrong dimensionality")

        # The following does *not* transpose a matrix defined as n x N=4**k
        # n is the number of independent files/samples, (fasta/fastq=>kdb) being assessed
        # N=4**k is the dimensionality of the vector, sparse or not, that makes up the perceived profile, the descriptors is the column dimension.
        # The names for the columns are taken as the basenamed filepath, with all extensions (substrings beginning with '.') stripped.

        if arguments.column_names is None:
            columns = list(map(lambda kdbrdr: os.path.basename(kdbrdr._filepath).split(".")[0], files))
        else:
            with open(arguments.column_names, 'r') as column_names:
                columns = [line.rstrip() for line in column_names]
        if len(columns) != len(files):
            raise RuntimeError("Number of column names {0} does not match number of input files {1}...".format(len(columns, len(files))))
        logger.debug("Shape: {0}".format(profiles.shape))
        logger.info("Converting arrays of k-mer counts into a pandas DataFrame...")
        df = pd.DataFrame(profiles)
        #columns = list(df.columns) # I hate this language, it's hateful.
        n = len(columns)
    elif len(arguments.input) == 1 and (os.path.splitext(arguments.input[0])[-1] == ".tsv" or os.path.splitext(arguments.input[0])[-1] == ".csv"):
        try:
            df = pd.read_csv(arguments.input[0], sep=arguments.delimiter)
        except pd.errors.EmptyDataError as e:
            logger.error(e)
            logger.error("Pandas error on DataFrame reading. Perhaps a null dataset being read?")
            sys.exit(1)
        profiles = np.array(df)
        columns = list(df.columns) # I'm sorry I ever made this line. Please forgive me.
        # This is just gratuitous code and language. I'm really really not sure what I want to express here.
        n = len(columns)
    elif len(arguments.input) == 1 and os.path.splitext(arguments.input[0])[-1] == ".kdb":
        logger.error("Not sure why you'd want a singular distance.")
        logger.error("kdb distance requires more than one .kdb file as positional inputs")
        sys.exit(1)
    elif len(arguments.input) == 1 and (arguments.input[0] == "STDIN" or arguments.input[0] == "/dev/stdin"):
        logger.info("Reading input as tsv/csv from STDIN")
        try:
            df = pd.read_csv(sys.stdin, sep=arguments.delimiter)
        except pd.errors.EmptyDataError as e:
            logger.error(e)
            logger.error("Pandas error on DataFrame reading. Perhaps a null dataset being read?")
            sys.exit(1)
        profiles = np.array(df)
        columns = list(df.columns)
        n = len(columns)
    else:
        logger.error("bin/kdb.distances() received {0} arguments as input, which were not supported.".format(len(arguments.input)))
        sys.exit(1)

    # Masthead stuff
    sys.stderr.write(config.DEFAULT_MASTHEAD)
    if logger.level == logging.DEBUG or logger.level == logging.INFO:
        sys.stderr.write(config.DEBUG_MASTHEAD)
    sys.stderr.write(config.DISTANCE_MASTHEAD)

    logger.info("Custom calculating a {0}x{0} '{1}' distance matrix...".format(n, arguments.metric))
    
    if arguments.metric in ["pearson", "correlation", "spearman", "EMD", "d2s"]:
        data = [['' for x in range(n)] for y in range(n)]
        for i in range(n):
            for j in range(n):
                logger.debug("Calculating the {0}x{1} cell".format(i, j))
                logger.info("Calculating {0} distance between {1}:({2}) and {3}:({4}) columns...".format(arguments.metric, columns[i], i, columns[j], j))

                if i == j:
                    logger.info("Was self, reverting to identity.")
                    data[i][j] = distance.identity[arguments.metric]
                elif i > j:
                    data[i][j] = None
                    logger.info("Refusing to recompute custom distance metric for symmetric matrix")
                elif i < j:
                    #print("Info: ixj {0}x{1}")
                    logger.debug("Info: ixj {0}x{1}".format(i, j))
                    
                    if arguments.metric == "correlation" or arguments.metric == "pearson":
                        logger.info("Computing custom correlation coefficient")
                        data[i][j] = distance.correlation(files[i], files[j], k=suggested_k, dtype=suggested_dtype)
                    elif arguments.metric == "euclidean":
                        logger.info("Computing custom euclidean distance")
                        data[i][j] = distance.euclidean(profiles[i], profiles[j])
                    elif arguments.metric == "spearman":
                        cor, pval = distance.spearman(profiles[i], profiles[j])
                        logger.info("Computing SciPy Spearman distance")
                        # FIXME! also print pval matrices
                        data[i][j] = cor
                    elif arguments.metric == "EMD":
                        data[i][j] = distance.EMD(profiles[i], profiles[j])
                    elif arguments.metric == "d2s":
                        data[i][j] = distance.d2s(profiles[i], profiles[j])
                    else:
                        logger.error("Other distances are not implemented yet")
                        sys.exit(1)
                # This double loop quickly identifies empty cells and sets the data correctly from the permutation above
        logger.info("Filling in symmetrical matrix...")
        for i in range(n):
            for j in range(n):
                if data[i][j] is None:
                    data[i][j] = data[j][i]
                    #logger.info("\n\n\nSwerve swerve\n\n\n")
        dist = np.array(data)
    else:
        dist = pdist(np.transpose(profiles), metric=arguments.metric)
        dist = squareform(dist)
        # if arguments.metric == "correlation":
        #     ones = np.ones(dist.shape, dtype="int64")
        #     dist = np.subtract(ones, dist)
    if dist.shape == (2,2):
        print(dist[0][1])
    else:
        df = pd.DataFrame(dist, columns=columns)
        
        ## FIXME: CUSTOM sorting code, not commiting to git repo
        #suffixes = [(int(x.split("_")[1]), i) for i, x in enumerate(column_names)] # A list of a 2-tuple of the correct sort order and the index
        #suffixes.sort(key=lambda x: x[0])
        #sorted_column_names = [column_names[s[1]] for s in suffixes]
        #df.sort_values(sorted_column_names)
        

        df.to_csv(sys.stdout, sep=arguments.output_delimiter, index=False)

def get_matrix(arguments):
    logging.getLogger('matplotlib.font_manager').disabled = True
    logging.getLogger('matplotlib').setLevel(logging.WARNING)

    import numpy as np
    import pandas as pd
    import matplotlib.pyplot as plt
    from sklearn.decomposition import PCA
    from sklearn.manifold import TSNE
    
    from kmerdb import fileutil, config


    try:
        np.dtype(arguments.dtype)
    except TypeError as e:
        logger.error(e)
        logger.error("kmerdb encountered a TypeError, halting.")
        raise argparse.ArgumentError("Dtype '{0}' is invalid.".format(arguments.dtype))
    
    if len(arguments.input) > 1:
        files = list(map(lambda f: fileutil.open(f, 'r'), arguments.input))
        dtypes = [x.dtype.name for x in files]
        suggested_dtype = dtypes[0]
        if arguments.k is None:
            arguments.k = files[0].k
        if not all(os.path.splitext(kdb)[-1] == ".kdb" for kdb in arguments.input):
            raise IOError("One or more parseable .kdb filepaths did not end in '.kdb'")
        ks = [kdbrdr.k for kdbrdr in files]
        suggested_k = ks[0]
        if not all(k == suggested_k for k in ks):
            logger.error("Files: {0}".format(files))
            logger.error("Choices of k: {0}".format(ks))
            logger.error("By default the inferred value of k is used when k is not specified at the command line, which was {0}".format(arguments.k))
            logger.error("Default choice of k, since not specified at the CLI is {0}".format(arguments.k))
            logger.error("Proceeding with k set to {0}".format(suggested_k))
            raise TypeError("One or more files did not have k set to be equal to {0}".format(arguments.k))
        elif not all(kdbrdr.dtype == arguments.dtype and arguments.dtype == kdbrdr.dtype for kdbrdr in files):
            raise TypeError("One or more files did not have the default dtype or the default dtype did not match one or more files...")
        profiles = np.transpose(np.array(list(map(lambda kdbrdr: kdbrdr.profile, files)), dtype=suggested_dtype))
        # The following does *not* transpose a matrix defined as n x N=4**k
        # n is the number of independent files/samples, (fasta/fastq=>kdb) being assessed
        # N=4**k is the dimensionality of the vector, sparse or not, that makes up the perceived profile, the descriptors is the column dimension.

        # The names for the columns are taken as the basenamed filepath, with all extensions (substrings beginning with '.') stripped.
        logger.info("============================================================")
        logger.info("Converting arrays of k-mer counts into a pandas DataFrame...")
        logger.info("============================================================")



        if arguments.column_names is None:
            columns = list(map(lambda kdbrdr: os.path.basename(kdbrdr._filepath).split(".")[0], files))
        else:
            with open(arguments.column_names, 'r') as column_names:
                # Expect column_names.txt to be one column name per line
                columns = [line.rstrip() for line in column_names]
        if len(columns) != len(files):
            raise RuntimeError("Number of column names {0} does not match number of input files {1}...".format(len(columns), len(files)))
        logger.debug("Shape: {0}".format(profiles.shape))
        expected = (4**suggested_k, len(columns))
        if profiles.shape != expected:
            logger.error("Expected shape: {0}".format(expected))
            logger.error("Actual shape: {0}".format(profiles.shape))
            raise RuntimeError("Raw profile shape (a Numpy array) doesn't match expected dimensions")

        df = pd.DataFrame(profiles, columns=columns)
        is_uint32 = all([x.dtype.name == 'uint32' for x in profiles])
        is_uint64 = all([x.dtype.name == 'uint64' for x in profiles])
        is_float32 = all([x.dtype.name == 'float32' for x in profiles])
        is_float64 = all([x.dtype.name == 'float64' for x in profiles])
        
        
    elif len(arguments.input) == 0 or (len(arguments.input) == 1 and (arguments.input[0] == "STDIN" or arguments.input[0] == "/dev/stdin")):
        logger.info("Reading input as tsv/csv from STDIN")
        try:
            df = pd.read_csv(sys.stdin, sep=arguments.delimiter)
        except pd.errors.EmptyDataError as e:
            logger.error(e)
            logger.error("Pandas error on DataFrame reading. Perhaps a null dataset being read?")
            sys.exit(1)
        profiles = np.array(df)
        columns = list(df.columns)
        n = len(columns)
        
    else:
        logger.error("bin/kdb.distances() received {0} arguments as input, which were not supported.".format(len(arguments.input)))
        sys.exit(1)

    # Masthead stuff
    sys.stderr.write(config.DEFAULT_MASTHEAD)
    if logger.level == logging.DEBUG or logger.level == logging.INFO:
        sys.stderr.write(config.DEBUG_MASTHEAD)
    sys.stderr.write(config.DISTANCE_MASTHEAD)

    logger.info("Calculating a {0}x{0} '{1}' distance matrix...".format(n, arguments.metric))
    
    if arguments.metric in ["spearman", "EMD", "d2s"]:
        data = [['' for x in range(n)] for y in range(n)]
        for i in range(n):
            for j in range(n):
                logger.info("Calculating {0} distance between {1} and {2}...".format(arguments.metric, column_names[i], column_names[j]))
                if i == j:
                    data[i][j] = distance.identity[arguments.metric]
                elif i > j:
                    data[i][j] = None
                elif i < j:
                    if arguments.metric == "correlation":
                        data[i][j] = distance.correlation(arguments.input[i], arguments.input[j])
                        # data[i][j] = distance.correlation(arguments.kdb[i], arguments.kdb[j])
                    elif arguments.metric == "euclidean":
                        data[i][j] = distance.euclidean(arguments.input[i], arguments.input[j])
                    elif arguments.metric == "spearman":
                        cor, pval = distance.spearman(profiles[i], profiles[j])
                        # FIXME! also print pval matrices
                        data[i][j] = cor
                    elif arguments.metric == "EMD":
                        data[i][j] = distance.EMD(profiles[i], profiles[j])
                    elif arguments.metric == "d2s":
                        data[i][j] = distance.d2s(profiles[i], profiles[j])
                    else:
                        logger.error("Other distances are not implemented yet")
                        sys.exit(1)
                # This double loop quickly identifies empty cells and sets the data correctly from the permutation above
        for i in range(n):
            for j in range(n):
                if data[i][j] is None:
                    data[i][j] = data[j][i]
        logger.info("Printing distance matrix...")
        logger.info(data)
        dist = np.array(data)
    else:
        dist = pdist(np.transpose(profiles), metric=arguments.metric)
        dist = squareform(dist)
        # if arguments.metric == "correlation":
        #     ones = np.ones(dist.shape, dtype="int64")
        #     dist = np.subtract(ones, dist)
    if dist.shape == (2,2):
        print(dist[0][1])
    else:
        df = pd.DataFrame(dist, columns=column_names)


        ## FIXME: CUSTOM sorting code, not commiting to git repo
        #suffixes = [(int(x.split("_")[1]), i) for i, x in enumerate(column_names)] # A list of a 2-tuple of the correct sort order and the index
        #suffixes.sort(key=lambda x: x[0])
        #sorted_column_names = [column_names[s[1]] for s in suffixes]
        #df.sort_values(sorted_column_names)
        

        df.to_csv(sys.stdout, sep=arguments.output_delimiter, index=False)

def get_matrix(arguments):
    logging.getLogger('matplotlib.font_manager').disabled = True
    logging.getLogger('matplotlib').setLevel(logging.WARNING)

    import numpy as np
    import pandas as pd
    import matplotlib.pyplot as plt
    from sklearn.decomposition import PCA
    from sklearn.manifold import TSNE
    
    from kmerdb import fileutil, config


    try:
        np.dtype(arguments.dtype)
    except TypeError as e:
        logger.error(e)
        logger.error("kmerdb encountered a TypeError, halting.")
        raise argparse.ArgumentError("Dtype '{0}' is invalid.".format(arguments.dtype))
    
    if len(arguments.input) > 1:
        files = list(map(lambda f: fileutil.open(f, 'r'), arguments.input))
        dtypes = [x.dtype.name for x in files]
        suggested_dtype = dtypes[0]
        if arguments.k is None:
            arguments.k = files[0].k
        if not all(os.path.splitext(kdb)[-1] == ".kdb" for kdb in arguments.input):
            raise IOError("One or more parseable .kdb filepaths did not end in '.kdb'")
        ks = [kdbrdr.k for kdbrdr in files]
        suggested_k = ks[0]
        if not all(k == suggested_k for k in ks):
            logger.error("Files: {0}".format(files))
            logger.error("Choices of k: {0}".format(ks))
            logger.error("By default the inferred value of k is used when k is not specified at the command line, which was {0}".format(arguments.k))
            logger.error("Default choice of k, since not specified at the CLI is {0}".format(arguments.k))
            logger.error("Proceeding with k set to {0}".format(suggested_k))
            raise TypeError("One or more files did not have k set to be equal to {0}".format(arguments.k))
        elif not all(kdbrdr.dtype == arguments.dtype and arguments.dtype == kdbrdr.dtype for kdbrdr in files):
            raise TypeError("One or more files did not have the default dtype or the default dtype did not match one or more files...")
        profiles = np.transpose(np.array(list(map(lambda kdbrdr: kdbrdr.slurp(dtype=suggested_dtype), files)), dtype=suggested_dtype))
        # The following does *not* transpose a matrix defined as n x N=4**k
        # n is the number of independent files/samples, (fasta/fastq=>kdb) being assessed
        # N=4**k is the dimensionality of the vector, sparse or not, that makes up the perceived profile, the descriptors is the column dimension.

        # The names for the columns are taken as the basenamed filepath, with all extensions (substrings beginning with '.') stripped.
        logger.info("============================================================")
        logger.info("Converting arrays of k-mer counts into a pandas DataFrame...")
        logger.info("============================================================")



        if arguments.column_names is None:
            columns = list(map(lambda kdbrdr: os.path.basename(kdbrdr._filepath).split(".")[0], files))
        else:
            with open(arguments.column_names, 'r') as column_names:
                # Expect column_names.txt to be one column name per line
                columns = [line.rstrip() for line in column_names]
        if len(columns) != len(files):
            raise RuntimeError("Number of column names {0} does not match number of input files {1}...".format(len(columns), len(files)))
        logger.debug("Shape: {0}".format(profiles.shape))
        expected = (4**suggested_k, len(columns))
        if profiles.shape != expected:
            logger.error("Expected shape: {0}".format(expected))
            logger.error("Actual shape: {0}".format(profiles.shape))
            raise RuntimeError("Raw profile shape (a Numpy array) doesn't match expected dimensions")

        #print(profiles.shape)
        logger.info("Created a matrix with the shape {0}".format(profiles.shape))
        #sys.exit(1)

        
        df = pd.DataFrame(profiles, columns=columns)
        is_uint32 = all([x.dtype.name == 'uint32' for x in profiles])
        is_uint64 = all([x.dtype.name == 'uint64' for x in profiles])
        is_float32 = all([x.dtype.name == 'float32' for x in profiles])
        is_float64 = all([x.dtype.name == 'float64' for x in profiles])
        
        
    elif len(arguments.input) == 0 or (len(arguments.input) == 1 and (arguments.input[0] == "STDIN" or arguments.input[0] == "/dev/stdin")):
        logger.info("Reading input as tsv/csv from STDIN")
        try:
            df = pd.read_csv(sys.stdin, sep=arguments.delimiter)
        except pd.errors.EmptyDataError as e:
            logger.error(e)
            logger.error("Pandas error on DataFrame reading. Perhaps a null dataset being read?")
            sys.exit(1)
        profiles = np.array(df)
        columns = list(df.columns)
        n = len(columns)
    elif len(arguments.input) == 1 and (os.path.splitext(arguments.input[0])[-1] == ".tsv" or os.path.splitext(arguments.input[0])[-1] == ".csv"):
        logger.info("Hidden: 1 argument. Reading input as tsv")
        try:
            df = pd.read_csv(arguments.input[0], sep=arguments.delimiter)
        except pd.errors.EmptyDataError as e:
            logger.error(e)
            logger.error("Pandas error on DataFrame reading. Perhaps a null dataset being read?")
            sys.exit(1)
        profiles = np.array(df)
        columns = list(df.columns)
        n = len(columns)
    elif len(arguments.input) == 1 and os.path.splitext(arguments.input)[-1] == ".kdb":
        logger.error("kdb distance requires more than one .kdb file as positional inputs")
        sys.exit(1)
    else:
        logger.error("bin/kdb.distances() received {0} arguments as input, which were not supported.".format(len(arguments.input)))
        sys.exit(1)

    # Masthead stuff
    sys.stderr.write(config.DEFAULT_MASTHEAD)
    if logger.level == logging.DEBUG or logger.level == logging.INFO:
        sys.stderr.write(config.DEBUG_MASTHEAD)
    sys.stderr.write(config.DISTANCE_MASTHEAD)

    logger.info("Calculating a {0}x{0} '{1}' distance matrix...".format(n, arguments.metric))
    
    if arguments.metric in ["spearman", "EMD", "d2s"]:
        data = [['' for x in range(n)] for y in range(n)]
        for i in range(n):
            for j in range(n):
                logger.info("Calculating {0} distance between {1} and {2}...".format(arguments.metric, columns[i], columns[j]))
                if i == j:
                    data[i][j] = distance.identity[arguments.metric]
                elif i > j:
                    data[i][j] = None
                elif i < j:
                    if arguments.metric == "correlation":
                        data[i][j] = distance.correlation(arguments.input[i], arguments.input[j])
                        # data[i][j] = distance.correlation(arguments.kdb[i], arguments.kdb[j])
                    elif arguments.metric == "euclidean":
                        data[i][j] = distance.euclidean(arguments.input[i], arguments.input[j])
                    elif arguments.metric == "spearman":
                        cor, pval = distance.spearman(profiles[i], profiles[j])
                        # FIXME! also print pval matrices
                        data[i][j] = cor
                    elif arguments.metric == "EMD":
                        data[i][j] = distance.EMD(profiles[i], profiles[j])
                    elif arguments.metric == "d2s":
                        data[i][j] = distance.d2s(profiles[i], profiles[j])
                    else:
                        logger.error("Other distances are not implemented yet")
                        sys.exit(1)
                # This double loop quickly identifies empty cells and sets the data correctly from the permutation above
        for i in range(n):
            for j in range(n):
                if data[i][j] is None:
                    data[i][j] = data[j][i]
        logger.info("Printing distance matrix...")
        logger.info(data)
        dist = np.array(data)
    else:
        dist = pdist(np.transpose(profiles), metric=arguments.metric)
        dist = squareform(dist)
        # if arguments.metric == "correlation":
        #     ones = np.ones(dist.shape, dtype="int64")
        #     dist = np.subtract(ones, dist)
    if dist.shape == (2,2):
        print(dist[0][1])
    else:
        df = pd.DataFrame(dist, columns=columns)


        ## FIXME: CUSTOM sorting code, not commiting to git repo
        #suffixes = [(int(x.split("_")[1]), i) for i, x in enumerate(column_names)] # A list of a 2-tuple of the correct sort order and the index
        #suffixes.sort(key=lambda x: x[0])
        #sorted_column_names = [column_names[s[1]] for s in suffixes]
        #df.sort_values(sorted_column_names)
        

        df.to_csv(sys.stdout, sep=arguments.output_delimiter, index=False)

def get_matrix(arguments):
    logging.getLogger('matplotlib.font_manager').disabled = True
    logging.getLogger('matplotlib').setLevel(logging.WARNING)

    import numpy as np
    import pandas as pd
    import matplotlib.pyplot as plt
    from sklearn.decomposition import PCA
    from sklearn.manifold import TSNE
    
    from kmerdb import fileutil, config


    try:
        np.dtype(arguments.dtype)
    except TypeError as e:
        logger.error(e)
        logger.error("kmerdb encountered a TypeError, halting.")
        raise argparse.ArgumentError("Dtype '{0}' is invalid.".format(arguments.dtype))
    
    if len(arguments.input) > 1:
        files = list(map(lambda f: fileutil.open(f, 'r'), arguments.input))
        if arguments.k is None:
            arguments.k = files[0].k
        if not all(os.path.splitext(kdb)[-1] == ".kdb" for kdb in arguments.input):
            raise IOError("One or more parseable .kdb filepaths did not end in '.kdb'")
        ks = [kdbrdr.k for kdbrdr in files]
        suggested_k = ks[0]
        if not all(k == suggested_k for k in ks):
            logger.error("Files: {0}".format(files))
            logger.error("Choices of k: {0}".format(ks))
            logger.error("By default the inferred value of k is used when k is not specified at the command line, which was {0}".format(arguments.k))
            logger.error("Default choice of k, since not specified at the CLI is {0}".format(arguments.k))
            logger.error("Proceeding with k set to {0}".format(suggested_k))
            raise TypeError("One or more files did not have k set to be equal to {0}".format(arguments.k))
        elif not all(kdbrdr.dtype == arguments.dtype and arguments.dtype == kdbrdr.dtype for kdbrdr in files):
            raise TypeError("One or more files did not have the default dtype or the default dtype did not match one or more files...")
        dtypes = [kdbrdr.dtype for kdbrdr in files]
        suggested_dtype = dtypes[0]
        if not all(kdbrdr.dtype == suggested_dtype for kdbrdr in files):
            raise TypeError("One of more files did not have dtype = {0}".format(suggested_dtype))
        data = [kdbrdr.profile for kdbrdr in files]
        pure_data = np.array(data, dtype=suggested_dtype)
        profiles = np.transpose(pure_data)
        # The following does *not* transpose a matrix defined as n x N=4**k
        # n is the number of independent files/samples, (fasta/fastq=>kdb) being assessed
        # N=4**k is the dimensionality of the vector, sparse or not, that makes up the perceived profile, the descriptors is the column dimension.

        # The names for the columns are taken as the basenamed filepath, with all extensions (substrings beginning with '.') stripped.
        logger.info("============================================================")
        logger.info("Converting arrays of k-mer counts into a pandas DataFrame...")
        logger.info("============================================================")



        if arguments.column_names is None:
            columns = list(map(lambda kdbrdr: os.path.basename(kdbrdr._filepath).split(".")[0], files))
        else:
            with open(arguments.column_names, 'r') as column_names:
                columns = [line.rstrip() for line in column_names]
        if len(columns) != len(files):
            raise RuntimeError("Number of column names {0} does not match number of input files {1}...".format(len(columns), len(files)))
        suggested_metadata = files[0].metadata
        expected, num_columns = 4**suggested_k, len(columns)
        if profiles.shape != (expected, num_columns):
            #logger.info("Time is money.")
            logger.error("Check shape self...")
            logger.error("Expected shape: {0} x {1}".format(expected, num_columns))
            logger.error("Actual shape: {0}".format(profiles.shape))
            raise RuntimeError("Raw profile shape (a NumPy array) doesn't match expected dimensions")
        df = pd.DataFrame(profiles, columns=columns)
        is_uint32 = False
        is_uint64 = False
        is_float32 = False
        is_float64 = False
        is_uint32 = all([x.dtype.name == 'uint32' for x in profiles])
        is_uint64 = all([x.dtype.name == 'uint64' for x in profiles])
        is_float32 = all([x.dtype.name == 'float32' for x in profiles])
        is_float64 = all([x.dtype.name == 'float64' for x in profiles])

    elif len(arguments.input) == 0 or (len(arguments.input) == 1 and (arguments.input[0] == "STDIN" or arguments.input[0] == "/dev/stdin")):
        logger.info("Reading input as tsv/csv from STDIN")
        try:
            df = pd.read_csv(sys.stdin, sep=arguments.delimiter)
        except pd.errors.EmptyDataError as e:
            logger.error(e)
            logger.error("Pandas error on DataFrame reading. Perhaps a null dataset being read?")
            sys.exit(1)
        columns = list(df.columns)
    elif len(arguments.input) == 1 and (os.path.splitext(arguments.input[0])[-1] == ".tsv" or os.path.splitext(arguments.input[0])[-1] == ".csv"):
        logger.info("Reading input file as tsv/csv")
        try:
            df = pd.read_csv(arguments.input[0], sep=arguments.delimiter)
        except pd.errors.EmptyDataError as e:
            logger.error(e)
            logger.error("Pandas error on DataFrame reading. Perhaps a null dataset being read?")
            sys.exit(1)
        columns = list(df.columns)
    else:
        logger.error(arguments)
        logger.error("bin/kdb.get_matrix() received {0} arguments as input, and this is not supported.".format(len(arguments.input)))
        logger.error(arguments.input)
        sys.exit(1)
    sys.stderr.write(config.DEFAULT_MASTHEAD)
    if logger.level == logging.DEBUG or logger.level == logging.INFO:
        sys.stderr.write(config.DEBUG_MASTHEAD)
    sys.stderr.write(config.MATRIX_MASTHEAD)


    final_df = None
    if arguments.method == "DESeq2":
        # if arguments.normalize_with == "ecopy":
        #     import ecopy as ep
        #     logger.info("Normalizing the DataFrame for sample size with ecopy...")
        #     normalized = ep.transform(df, method="normalize", axis=0)
        # # Normalizing across the 0th axis will normalize across the row meaning between samples. To arrive at this I checked the column sums of both to be sure I understood what was being normalized.
        # # We don't necessarily need to standardize, and not standardizing will make rationalizations or sensitivity to the real number range more clear.
        # if arguments.normalize_with == "DESeq2":

        logger.info("Normalizing the DataFrame with DESeq2 NB-compatible count normalization via rpy2...")
        '''
        The following was shown to me by the following medium post.
        https://w1ndy.medium.com/calling-r-libraries-from-python-5ffbf3c3e5a8
        '''
        from rpy2.robjects.packages import importr, PackageNotInstalledError
        from rpy2.robjects import FloatVector, DataFrame, r, pandas2ri, Formula
        pandas2ri.activate()
        try:
            #fitdistrplus = importr('fitdistrplus')
            deseq = importr('DESeq2')
            count_matrix = pandas2ri.py2rpy(df)
            ncol = r['ncol']
            seq = r['seq']
            colData = pd.DataFrame(np.transpose(columns), columns=["Species"])
            logger.info("colData:\n{0}".format(colData))
                
            dds = r['DESeqDataSetFromMatrix'](count_matrix, colData, Formula('~ Species'))
            logger.info("DESeq dataset:\n{0}".format(str(dds)))
            dds = r['estimateSizeFactors'](dds)
                
            normalized = r['counts'](dds, normalized = True)
            if not arguments.no_normalized_ints:
                normalized = np.array(np.rint(normalized), dtype="int64")
            normalized = pd.DataFrame(normalized, columns=columns)
            logger.debug(normalized)
            logger.info("Normalized matrix shape: {0}".format(normalized.shape))

            #     if arguments.method == "Unnormalized":
            #         cf = r['descdist'](r_dataframe, discrete=True)
            #     else:
            #         cf = r['descdist'](r_dataframe, discrete=False)
        except PackageNotInstalledError as e:
            logger.error(e)
            logger.error("One or more R packages were not installed. Please see the install documentation about installing the associated R packages")
            logger.error("Use -vv for suggested installation instructions.")
            sys.exit(1)
        #normalized.to_csv(sys.stdout, sep=arguments.delimiter, index=arguments.with_index)

        # logger.error("False ending of Normalized")
        # logger.debug("final_df should be set as normalized")
        # sys.exit(1)
        final_df = normalized
    elif arguments.method == "pass":
        #df.to_csv(sys.stdout, sep=arguments.delimiter, index=arguments.with_index)
        final_df = df
    elif arguments.method == "Frequency":
        k = suggested_metadata['k']
        total_kmers = suggested_metadata["total_kmers"]
        final_df = df.div(total_kmers)
    elif arguments.method == "PCA":
        # This method is actually dimensionality reduction via SVD, and this process is used during principal components analysis.
        # We generate the elbow graph in this step if the required dimensionality parameter '-n' is not supplied.
        # In this case we assume they have not provided the parameter because they'd like to use the so-called 'auto-detection'
        
        # I've specified a feature called 'auto-detect' such that the elbow graph is produced
        # they supply the number of dimensions they'd like after viewing the graph,
        # And in this way by viewing the graph, the 'correct' number of dimensions is attained.
        #
        #################################
        # PCA dimensionality reduction
        #################################
    
        logger.info("Performing preliminary dimensionality reduction to the MLE")
        pca = PCA()#n_components="mle", svd_solver='auto')
        pca.fit(np.transpose(df)) # PCA of the normalized matrix or its transpose?
        # We want the number of k-mers, the number of features reduced, so we transpose the original matrix


        plt.plot(range(1, len(pca.explained_variance_ratio_)+1), pca.explained_variance_ratio_.cumsum(), marker='o', linestyle="--")
        plt.title("Explained variance by components")
        plt.xlabel("Number of components")
        plt.ylabel("Cumulative explained variance")
        plt.savefig(config.pca_variance_fig_filepath)

        if arguments.n is not None:
            logger.info("Using selected PCA dimensionality to reduce the transpose matrix/DataFrame again for use in 'kdb kmeans'")
            pca = PCA(n_components=arguments.n)
            pca.fit(np.transpose(df))
            #logger.debug("Explained variances: {0}".format(pca.explained_variance_ratio_))
            #logger.debug("Log-likelihoods: {0}".format(pca.score_samples(normalized)))
            #logger.debug("Overall log-likelihood of all samples: {0}".format(pca.score(normalized)))
            #logger.debug("MLE estimate of components for dimensionality reduction produced this shape: {0}".format(pca.components_.shape))

            score_matrix = pca.transform(np.transpose(df))
            score_df = pd.DataFrame(np.transpose(score_matrix), columns=columns)

            #score_df.to_csv(sys.stdout, sep=arguments.delimiter, index=arguments.with_index)
            final_df = score_df
        else:
            logger.warning("You must look at '{0}' to decide on the choice of '-n' for specify when generating the dimensionally reduced matrix for clustering function".format(config.pca_variance_fig_filepath))
            sys.exit(1)
        #logger.debug("Components:\n{0}".format(pca.components_))
        # The shape for my first run was 10x11 and I have 11 samples. So I decided to not transpose this matrix to simply add it to a DataFrame and rename the samples.
        #df = pd.DataFrame(pca.components_, columns=list(map(lambda kdbrdr: os.path.splitext(os.path.basename(kdbrdr._filepath))[0], files)))
    elif arguments.method == "tSNE":
        '''
        In t-SNE, perplexity or k is defined as 2^S, where S is the Shannon entropy of the conditional probability distribution.
        '''
        if arguments.n is None:
            raise TypeError("'kdb matrix tSNE' requires a keyword argument '-n' equal to the number of components of the subspace for tSNE to project the data into. A choice of 2 is recommended")
        tsne = TSNE(n_components=arguments.n, perplexity=arguments.perplexity).fit_transform(np.transpose(df))
        tsne_df = pd.DataFrame(np.transpose(tsne), columns=columns)
        #tsne_df.to_csv(sys.stdout, sep=arguments.delimiter, index=arguments.with_index)
        final_df = tsne_df

    ## FIXME: CUSTOM SORTING. WILL NOT COMMIT TO GIT REPO
    #suffixes = [(int(x.split("_")[1]), i) for i, x in enumerate(column_names)] # A list of a 2-tuple of the correct sort order and the index
    #suffixes.sort(key=lambda x: x[0])
    #sorted_column_names = [column_names[s[1]] for s in suffixes]
    #final_df.sort_values(sorted_column_names)

    final_df.to_csv(sys.stdout, sep=arguments.output_delimiter, index=arguments.with_index)
    logger.info("Done printing {0} matrix to STDOUT".format(arguments.method))

    #logger.info("Beginning distribution analysis in R...")
    #logger.warn("Not implemented in Python...")
    
    sys.stderr.write(config.DONE)



def kmeans(arguments):
    logging.getLogger('matplotlib.font_manager').disabled = True
    logging.getLogger('matplotlib').setLevel(logging.WARNING)
    from io import TextIOWrapper
    import numpy as np
    import pandas as pd
    import matplotlib.pyplot as plt
    from sklearn import metrics
    from sklearn.cluster import KMeans
    from sklearn.pipeline import make_pipeline
    from sklearn.preprocessing import StandardScaler

    from kmerdb import config

    
    if isinstance(arguments.input, argparse.FileType) or isinstance(arguments.input, TextIOWrapper):
        df = pd.read_csv(arguments.input, sep=arguments.delimiter)
        column_names = df.columns
        df = df.transpose()
        #df.set_index('index')
    elif arguments.input is None or arguments.input == "/dev/stdin" or arguments.input == "STDIN":
        df = pd.read_csv(sys.stdin, sep=arguments.delimiter)
        column_names = df.columns
        df = df.transpose()
        #df.set_index('index')
    else:
        logger.error("An unknown IO type was detected in bin/kdb.cluster()")
        sys.exit(1)

    sys.stderr.write(config.DEFAULT_MASTHEAD) # Not working
    if logger.level == logging.DEBUG or logger.level == logging.INFO: # Not working
        sys.stderr.write(config.DEBUG_MASTHEAD)
    sys.stderr.write(config.KMEANS_MASTHEAD) # Not working

        
    num_samples, num_features = df.shape
    logger.info("Input DataFrame shape: {0}".format(df.shape))
    t0 = time.time()

    
    wcss = []
    for i in range(1, num_samples+1):
        kmeans_pca = KMeans(n_clusters = i, init="k-means++", random_state=42)
        kmeans_pca.fit(df)
        wcss.append(kmeans_pca.inertia_)

    plt.plot(range(1, num_samples+1), wcss, marker='o', linestyle='--')
    plt.xlabel("Number of Clusters")
    plt.ylabel("WCSS")
    plt.title("K-means with PCA")
    plt.savefig(config.kmeans_elbow_graph_fig_filepath)
    sys.stderr.write("="*42 + "\n\n")
    sys.stderr.write(" " * 10 + "The range 1:{0} was swept to find an elbow for the clustering\n".format(num_samples))
    sys.stderr.write("The figure was written to '{0}'\n\n".format(config.kmeans_elbow_graph_fig_filepath))
    sys.stderr.write("="*42 + "\n\n")

    
    if arguments.k is not None and arguments.method == "sklearn":
        final_kmeans = KMeans(n_clusters = arguments.k, init="k-means++", random_state=42)
        final_kmeans.fit(df)
        labels = final_kmeans.labels_
        df.insert(0, "Cluster", labels)
        #labels = None
        # final_df.replace({"Cluster": {
        #     1: "One"
        # }})
        df.to_csv(arguments.output, sep=arguments.output_delimiter)
        colnames = list(df.columns)
        fig, ax = plt.subplots()
        scatter = ax.scatter(df.iloc[:, 1], df.iloc[:, 2], c=df.iloc[:, 0])
        legend = ax.legend(*scatter.legend_elements(), loc="upper right", title="Cluster")

        for x,y,z in zip(np.array(df.iloc[:, 1]), np.array(df.iloc[:, 2]), column_names):
            sys.stderr.write("{0}\t{1}\t{2}".format(x,y,z))
            ax.annotate(z, xy=(x, y), xycoords='data', xytext=(0, -45), textcoords='offset points', arrowprops=dict(facecolor='black', shrink=0.05), horizontalalignment='left', verticalalignment='bottom')
        ax.set_xlabel("Dim1")
        ax.set_ylabel("Dim2")
        ax.set_title("K-means clustering")
        ax.add_artist(legend)
        ax.grid(True)


        plt.savefig(config.kmeans_clustering_fig_filepath)
        
        #centroids = kmeans.cluster_centers_
        logger.info(list(column_names))
        logger.info(labels)
    
        '''
        Clustering score quality estimation from 
        https://scikit-learn.org/stable/auto_examples/cluster/plot_kmeans_digits.html#define-our-evaluation-benchmark
    
        '''
        name = "K-means metrics"
        fit_time = time.time() - t0
        results = [name, fit_time, final_kmeans.inertia_]

        clustering_metrics = [
            metrics.homogeneity_score,
            metrics.completeness_score,
            metrics.v_measure_score,
            metrics.adjusted_rand_score,
            metrics.adjusted_mutual_info_score,
        ]
        
        results += [m(labels, final_kmeans.labels_) for m in clustering_metrics]

        results += [metrics.silhouette_score(df, final_kmeans.labels_, metric="euclidean", sample_size=300,)]

        formatter_result = ("\n{:9s}\t{:.3f}s\t{:.0f}\t{:.3f}\t{:.3f}"
                            "\t{:.3f}\t{:.3f}\t{:.3f}\t{:.3f}\n\n\n")
        sys.stderr.write("\n" + "\t".join(["Name", "Total_time", "Inertia", "Homogeneity_score", "Completeness_score", "V_measure_score", "Adj_rand_score", "Adj_mutual_info_score", "Silhouette"]) + "\n")
        sys.stderr.write(formatter_result.format(*results))

        #kmer_pca = pca(df, scale=False) # It's already been normalized (scaled) in get_matrix()
        #print(kmer.summary_rot())


    elif arguments.k is not None and arguments.method == "Biopython":
        from Bio import Cluster

        clust, error, nfound = Cluster.kcluster(df, nclusters=arguments.k, dist=arguments.distance)

        labels = list(clust)

        
        df.insert(0, "Cluster", labels)
        #labels = None
        # final_df.replace({"Cluster": {
        #     1: "One"
        # }})

        df.to_csv(arguments.output, sep=arguments.output_delimiter)
        fig, ax = plt.subplots()
        scatter = ax.scatter(df.iloc[:, 1], df.iloc[:, 2], c=df.iloc[:, 0])
        legend = ax.legend(*scatter.legend_elements(), loc="upper right", title="Cluster")
        for x,y,z in zip(np.array(df.iloc[:, 1]), np.array(df.iloc[:, 2]), column_names):
            sys.stderr.write("{0}\t{1}\t{2}".format(x,y,z))
            ax.annotate(z, xy=(x, y), xycoords='data', xytext=(0, -45), textcoords='offset points', arrowprops=dict(facecolor='black', shrink=0.05), horizontalalignment='left', verticalalignment='bottom')

        
        ax.set_xlabel("Dim1")
        ax.set_ylabel("Dim2")
        ax.set_title("K-means clustering")
        ax.add_artist(legend)
        ax.grid(True)
        plt.savefig(config.kmeans_clustering_fig_filepath)
        
        #centroids = kmeans.cluster_centers_
        logger.info(list(column_names))
        logger.info(labels)
        
    sys.stderr.write("Generated {0} clusters projected onto reduced dimension 1 and 2 of the input dataset\n".format(arguments.k))
    sys.stderr.write("The figure was written to {0}\n\n".format(config.kmeans_clustering_fig_filepath))

    logger.info("The annotated sample matrix can be printed by specifying an output files with [ -o|--output OUTFILE ]")
    
    sys.stderr.write(config.DONE)



def hierarchical(arguments):
    """
    Thanks to https://www.analyticsvidhya.com/blog/2019/05/beginners-guide-hierarchical-clustering/
    for the inspiration
    """
    logging.getLogger('matplotlib.font_manager').disabled = True
    logging.getLogger('matplotlib').setLevel(logging.WARNING)
    from io import TextIOWrapper
    import numpy as np
    import pandas as pd
    import matplotlib.pyplot as plt
    import scipy.cluster.hierarchy as shc

    from Bio.Cluster import treecluster
    
    from kmerdb import config


    
    if isinstance(arguments.input, argparse.FileType) or isinstance(arguments.input, TextIOWrapper):
        df = pd.read_csv(arguments.input, sep=arguments.delimiter)
        column_names = list(df.columns)
        df = df.transpose()
    elif arguments.input is None or arguments.input == "STDIN" or arguments.input == "/dev/stdin":
        df = pd.read_csv(sys.stdin, sep=arguments.delimiter)
        column_names = list(df.columns)
        df = df.transpose()
    else:
        logger.error("An unknown IO type was detected in bin/kdb.cluster()")
        sys.exit(1)

    sys.stderr.write(config.DEFAULT_MASTHEAD)
    if logger.level == logging.DEBUG or logger.level == logging.INFO:
        sys.stderr.write(config.DEBUG_MASTHEAD)
    sys.stderr.write(config.HIERARCHICAL_MASTHEAD)
    
    num_samples, num_features = df.shape
    logger.info("Input DataFrame shape: {0}".format(df.shape))

    #tree = treecluster(distancematrix=np.array(df)) # Can this be exported to Newick?
    plt.figure(figsize=(16, 9)) # figsize=(10, 7)
    plt.title("Dendrogram")
    dend = shc.dendrogram(shc.linkage(df, method=arguments.method), labels=column_names, leaf_rotation=90)
    #plt.show()
    plt.savefig(config.hierarchical_clustering_dendrogram_fig_filepath)

    sys.stderr.write("Saving the matplotlib dendrogram to '{0}'...".format(config.hierarchical_clustering_dendrogram_fig_filepath))

    data = np.array(df).tolist()

    n = len(data)
    m = len(data[0])
    for i in range(n):
        for j in range(m):
            if i < j:
                data[i].pop(-1)
    #print(data)
    
    from Bio.Phylo.TreeConstruction import DistanceMatrix, DistanceTreeConstructor
    from Bio import Phylo
    dm = DistanceMatrix(column_names, data)
    constructor = DistanceTreeConstructor()
    upgmatree = constructor.upgma(dm)
    print(upgmatree)
    Phylo.draw(upgmatree, branch_labels=lambda c: c.branch_length)
    Phylo.write(upgmatree, config.spearman_upgma_tree_phy, "phyloxml")
    sys.stderr.write("Saving the phylip format tree to '{0}'".format(config.spearman_upgma_tree_phy))
    sys.stderr.write(config.DONE)

# def rarefy(arguments):
#     logging.getLogger('matplotlib.font_manager').disabled = True
#     logging.getLogger('matplotlib').setLevel(logging.WARNING)
#     from io import TextIOWrapper
#     import numpy as np
#     import pandas as pd
#     import matplotlib.pyplot as plt

    
#     from ecopy.diversity import rarefy
#     import ecopy as ep
#     from kdb import config


#     if isinstance(arguments.input, argparse.FileType) or isinstance(arguments.input, TextIOWrapper):
#         df = pd.read_csv(arguments.input, sep=arguments.delimiter)
#         column_names = df.columns
#         df = df.transpose()
#     elif arguments.input is None:
#         df = pd.read_csv(sys.stdin, sep=arguments.delimiter)
#         column_names = df.columns
#         df = df.transpose()
#     else:
#         logger.error("An unknown IO type was detected in bin/kdb.cluster()")
#         sys.exit(1)

#     sys.stderr.write(config.DEFAULT_MASTHEAD)
#     if logger.level == logging.DEBUG:
#         sys.stderr.write(config.DEBUG_MASTHEAD)
#     sys.stderr.write(config.RAREFY_MASTHEAD)


#     num_samples, num_features = df.shape
#     t0 = time.time()

#     logger.warning("Multiple files: Untested behavior with ecopy.diversity.rarefy")
#     test = rarefy(df, config.ecopy_rarefaction_fig_filepath, 'rarecurve')
#     #arguments.output.write(test)
#     if arguments.output is None:
#         print(test)
#     else:
#         arguments.output.write(test)

#     sys.stderr.write(config.DONE)
        
        

def header(arguments):
    from kmerdb import fileutil, config, util

    if os.path.splitext(arguments.kdb)[-1] != ".kdb":
        raise IOError("Viewable .kdb filepath does not end in '.kdb'")

    with fileutil.open(arguments.kdb, mode='r') as kdb:
        if kdb.metadata["version"] != config.VERSION:
            logger.warning("KDB version is out of date, may be incompatible with current KDBReader class")
        if arguments.json:
            print(dict(kdb.metadata))
        else:
            yaml.add_representer(OrderedDict, util.represent_ordereddict)
            print(yaml.dump(kdb.metadata))
            print(config.header_delimiter)
            
def view(arguments):
    import numpy as np
    from kmerdb import fileutil, config, util
    import json
    metadata = None

    def get_header(line, header):
        """
        A little helper recurrence function for grabbing the additions to the header.
        I don't really know if this is germane... Untested  function.
        """
        

        if line.rstrip() != config.end_header_line:
            header += line
            return header
        else:
            header_dict = yaml.safe_load(header)
            if type(header_dict) is not dict:
                logger.debug("Tried to parse:\n{0}\n".format(header))
                raise ValueError("Could not parse YAML formatted header")
            else:
                return header_dict

    assert type(arguments.kdb_in) is str, "kdb_in must be a str"
    if os.path.splitext(arguments.kdb_in)[-1] != ".kdb": # A filepath with invalid suffix
        raise IOError("Viewable .kdb filepath does not end in '.kdb'")
    elif not os.path.exists(arguments.kdb_in):
        raise IOError("Viewable .kdb filepath '{0}' does not exist on the filesystem".format(arguments.kdb_in))
    with fileutil.open(arguments.kdb_in, mode='r', sort=arguments.sorted, slurp=True) as kdb_in:
        metadata = kdb_in.metadata
        kmer_ids_dtype = metadata["kmer_ids_dtype"]
        if metadata["version"] != config.VERSION:
            logger.warning("KDB version is out of date, may be incompatible with current KDBReader class")
        if arguments.kdb_out is None or (arguments.kdb_out == "/dev/stdout" or arguments.kdb_out == "STDOUT"): # Write to stdout, uncompressed
            if arguments.header:
                yaml.add_representer(OrderedDict, util.represent_ordereddict)
                print(yaml.dump(metadata, sort_keys=False))
                print(config.header_delimiter)
        logger.info("Reading from file...")
        logger.debug("I cut off the json-formatted unstructured column for the main view.")
        logger.debug(kdb_in.profile)
        try:
            if arguments.sorted and metadata["sorted"] is True:
                kmer_ids_sorted_by_count = np.lexsort((kdb_in.counts, kdb_in.kmer_ids))
                for i, idx in enumerate(kmer_ids_sorted_by_count):
                    p = kdb_in.profile[i]
                    kmer_id = kdb_in.kmer_ids[kdb_in.profile[i]]
                    logger.debug("{0}\t{1}\t{2}\t{3}\t{4}".format(i, p, kmer_id, kdb_in.counts[kmer_id], kdb_in.frequencies[kmer_id]))
                    logger.debug(type(i))
                    logger.debug(type(idx))
                    print("{0}\t{1}\t{2}\t{3}\t{4}".format(i, kmer_id, p, kdb_in.counts[kmer_id], kdb_in.frequencies[kmer_id]))
            else:
                for i, idx in enumerate(kdb_in.profile):
                    p = kdb_in.profile[i]
                    kmer_id = kdb_in.kmer_ids[kdb_in.profile[i]]
                    logger.debug("The row in the file should follow this order:")
                    logger.debug("{0}\t{1}\t{2}\t{3}\t{4}".format(i, p, kmer_id, kdb_in.counts[kmer_id], kdb_in.frequencies[kmer_id]))
                    logger.debug(type(i))
                    logger.debug(type(idx))
                    try:
                        assert idx == kdb_in.profile[i], "The row/kmer id we're focusing is not found in the profile under this index"
                        #assert i == kdb_in.profile[i], "The k-mer row did not match the absorbed profile"
                        #assert p == i, "The row/kmer id we're focusing on is not found in this sort order. Column 1 is always row index, column2 is always profile sort order, then k-mer id, counts etc."
                        if p != i and metadata["sorted"] is True:
                            logger.warning("We are dealing with a sorted file... reading into memory unsorted")
                        else:
                            logger.warning("We are dealing with a sorted file... reading into memory unsorted")
                    except AssertionError as e:
                        logger.error("Row enumeration: {0}".format(i))
                        logger.error("Profile line: {0}".format(kdb_in.profile[i]))
                        logger.error("Kmer id: {0}".format(kmer_id))
                        logger.error("Count: {0}".format(kdb_in.counts[idx]))
                        logger.error(e)
                        raise e
                        #pass
                    logger.debug("{0} line:".format(i))
                    logger.debug("=== = = = ======= =  =  =  =  =  = |")

                    print("{0}\t{1}\t{2}\t{3}\t{4}".format(i, kmer_id, p, kdb_in.counts[kmer_id], kdb_in.frequencies[kmer_id]))
                # I don't think anyone cares about the graph representation.
                # I don't think this actually matters because I can't figure out what the next data structure is.
                # Is it a Cypher query and creation node set?
                # I need to demonstrate a capacity for graph based learning.
                # (:-|X) The dread pirate roberts got me.
                # :)
        except BrokenPipeError as e:
            logger.error(e)
            raise e
    if arguments.kdb_out is not None and arguments.compress: # Can't yet write compressed to stdout
        logger.error("Can't write kdb to stdout! We need to use a Bio.bgzf filehandle.")
        sys.exit(1)
    elif arguments.kdb_out is not None and type(arguments.kdb_out) is not str:
        raise ValueError("Cannot write a file to an argument that isn't a string")
    elif arguments.kdb_out is not None and os.path.exists(arguments.kdb_out):
        logger.warning("Overwriting '{0}'...".format(arguments.kdb_out))
    elif arguments.kdb_out is not None and not os.path.exists(arguments.kdb_out):
        logger.debug("Creating '{0}'...".format(arguments.kdb_out))
    if arguments.kdb_out is not None:
        raise RuntimeError("Unsupported")
        with fileutil.open(arguments.kdb_in, 'r', dtype=suggested_dtype, sort=arguments.sorted, slurp=True) as kdb_in:
            with fileutil.open(arguments.kdb_out, metadata=metadata, mode='w') as kdb_out:
                try:
                    line = None
                    for line in kdb_in:
                        line = line.rstrip()
                        kmer_id, count, kmer_metadata = line.line.rstrip().split("\t")
                        kdb_out.write("{0}\t{1}\t{2}\n".format(kmer_id, count, kmer_metadata))
                    
                except StopIteration as e:
                    logger.error(e)
                    raise e
                finally:
                    #kdb_out._write_block(kdb_out._buffer)
                    #kdb_out._handle.flush()
                    #kdb_out._handle.close()
                    sys.stderr.write(config.DONE)

            
def profile(arguments):
    from multiprocessing import Pool
    import json
    import time
    import numpy as np
    from kmerdb import parse, fileutil, kmer, util
    from kmerdb.config import VERSION

    logger.debug("Printing entire CLI argparse option Namespace...")
    logger.debug(arguments)

    # The number of specified processors should be available
    logger.debug("Validating processor count for parallelization...")
    if arguments.parallel <= 0 or arguments.parallel > cpu_count()+1:
        raise argparse.ArgumentError("-p, --parallel must be a valid processor count.")

    # The extension should be .kdb because I said so.
    logger.info("Checking extension of output file...")
    if os.path.splitext(arguments.kdb)[-1] != ".kdb":
        raise IOError("Destination .kdb filepath does not end in '.kdb'")




    file_metadata = []
    total_kmers = 4**arguments.k # Dimensionality of k-mer profile
    N = total_kmers

    logger.info("Parsing {0} sequence files to generate a composite k-mer profile...".format(len(list(arguments.seqfile))))
    nullomers = set()
    # Initialize arrays
    

    infile = parse.Parseable(arguments) #

    
    logger.info("Processing {0} fasta/fastq files across {1} processors...".format(len(list(arguments.seqfile)), arguments.parallel))
    logger.debug("Parallel (if specified) mapping the kmerdb.parse.parsefile() method to the seqfile iterable")
    logger.debug("In other words, running the kmerdb.parse.parsefile() method on each file specified via the CLI")
    if arguments.parallel > 1:
        with Pool(processes=arguments.parallel) as pool:
            data = pool.map(infile.parsefile, arguments.seqfile)
    else:
        data = list(map(infile.parsefile, arguments.seqfile))

    # 'data' is now a list of 4-tuples
    # Each 4-tuple represents a single file
    # (counts<numpy.array>, header_dictionary<dict>, nullomers<list>, all_kmer_metadata<list>)

    # Construct a final_counts array for the composite profile across all inputs
    logger.debug("Initializing large list for extended metadata")
    all_kmer_metadata = list([] for x in range(total_kmers)) if arguments.all_metadata else None
    logger.debug("Allocation completed")
    counts = np.zeros(N, dtype="uint64")
    # Complete collating of counts across files
    # This technically uses 1 more arrray than necessary 'final_counts' but its okay
    logger.info("Summing counts from individual fasta/fastq files into a composite profile...")
    for d in data:
        counts = counts + d[0]
        if arguments.all_metadata:
            logger.info("\n\nMerging metadata from all files...\n\n")
            all_kmer_metadata = util.merge_metadata_lists(arguments.k, all_kmer_metadata, d[3])

    sys.stderr.write("\n\n\tCompleted summation and metadata aggregation across all inputs...\n\n")
    unique_kmers = int(np.count_nonzero(counts))
    total_nullomers = total_kmers - unique_kmers
    all_observed_kmers = int(np.sum(counts))

    logger.info("Initial counting process complete, creating BGZF format file (.kdb)...")
    logger.info("Formatting master metadata dictionary...")

    metadata=OrderedDict({
        "version": VERSION,
        "metadata_blocks": 1,
        "k": arguments.k,
        "total_kmers": all_observed_kmers,
        "unique_kmers": unique_kmers,
        "unique_nullomers": total_nullomers,
        "metadata": arguments.all_metadata,
        "sorted": arguments.sorted,
        "kmer_ids_dtype": "uint64",
        "profile_dtype": "uint64",
        "count_dtype": "uint64",
        "frequencies_dtype": "float64",
        "tags": [],
        "files": [d[1] for d in data]
    })
        
    try:
        np.dtype(metadata["kmer_ids_dtype"])
        np.dtype(metadata["profile_dtype"])
        np.dtype(metadata["count_dtype"])
        np.dtype(metadata["frequencies_dtype"])
    except TypeError as e:
        logger.error(e)
        logger.error("kmerdb encountered a type error and needs to exit")
        raise TypeError("Incorrect dtype.")

    logger.debug("Initializing Numpy array of {0} uint zeroes for the final composite profile...".format(total_kmers))
    kmer_ids = np.array(range(N), dtype=metadata["kmer_ids_dtype"])
    profile = np.array(range(N), dtype=metadata["profile_dtype"])
    counts = np.array(counts, dtype=metadata["count_dtype"])
    frequencies = np.divide(counts, metadata["total_kmers"])
    logger.info("Initialization of profile completed, using approximately {0} bytes per profile".format(counts.nbytes))    


    


    logger.info("Collapsing the k-mer counts across the various input files into the final kdb file '{0}'".format(arguments.kdb)) 
    kdb_out = fileutil.open(arguments.kdb, 'wb', metadata=metadata)
    try:
        sys.stderr.write("\n\nWriting outputs to {0}...\n\n".format(arguments.kdb))


        # Numpy indexing is trash
        #kmer_ids = np.zeros(total_kmers, dtype=metadata["kmer_ids_dtype"])
        #profile = np.zeros(total_kmers, dtype=metadata["profile_dtype"])
        #counts = np.zeros(total_kmers, dtype=metadata["count_dtype"])
        #frequencies = np.zeros(total_kmers, dtype=metadata["frequencies_dtype"])
        all_metadata = []
        if arguments.all_metadata:

            if arguments.sorted:
                list_of_duples = list(zip(kmer_ids, counts))
                logger.debug(list_of_duples[0:3])
                kmer_ids_sorted_by_count = np.lexsort(list_of_duples)
                for i, idx in enumerate(kmer_ids_sorted_by_count):
                    seq = kmer.id_to_kmer(idx, arguments.k)
                    kmer_metadata = kmer.neighbors(seq, arguments.k) # metadata is initialized by the neighbors
                    reads = []
                    starts = []
                    reverses = []
                    frequency = float(count)/N
                    for read, start, reverse in all_kmer_metadata[idx]:
                        reads.append(read)
                        starts.append(start)
                        reverses.append(reverse)
                        kmer_metadata["reads"] = reads
                        kmer_metadata["starts"] = starts
                        kmer_metadata["reverses"] = reverses
                    counts[idx] = counts[idx]
                    frequencies[idx] = frequencies[idx]
                    all_metadata.append(kmer_metadata)
                    print("{0}\t{1}\t{2}\t{3}\t{4}".format(i, idx, kmer_ids[i], counts[idx], frequencies[idx]))
                    kdb_out.write("{0}\t{1}\t{2}\t{3}\t{4}\t{5}\n".format(i, idx, kmer_ids[i], counts[idx], frequencies[idx], json.dumps(kmer_metadata)))
            else:
                for i, idx in enumerate(kmer_ids):
                    kmer_id = int(idx)
                    seq = kmer.id_to_kmer(kmer_id, arguments.k)
                    logger.info("{0}\t{1}\t{2}\t{3}\t{4}".format(idx, idx, kmer_ids[i], counts[kmer_id], frequencies[kmer_id]))

                    kmer_metadata = kmer.neighbors(seq, arguments.k)
                    all_metadata.append(kmer_metadata)
                    counts[idx] = counts[idx]
                    frequencies[idx] = frequencies[idx]
                    print("{0}\t{1}\t{2}\t{3}\t{4}".format(i, idx, kmer_ids[i], counts[idx], frequencies[idx]))
                    kdb_out.write("{0}\t{1}\t{2}\t{3}\t{4}\t{5}\n".format(i, idx, kmer_ids[i], counts[i], frequencies[i], json.dumps(kmer_metadata)))
        else:
            if arguments.sorted:
                duple_of_arrays = (kmer_ids, counts)
                kmer_ids_sorted_by_count = np.lexsort(duple_of_arrays)
                logger.debug("K-mer id sort shape: {0}".format(len(kmer_ids_sorted_by_count)))
                for i, idx in enumerate(kmer_ids_sorted_by_count):

                    kmer_id = int(kmer_ids[idx])
                    logger.info("{0}\t{1}\t{2}\t{3}\t{4}".format(i, idx, kmer_ids[i], counts[idx], frequencies[idx]))
                    seq = kmer.id_to_kmer(kmer_id, arguments.k)
                    kmer_metadata = kmer.neighbors(seq, arguments.k)
                    all_metadata.append(kmer_metadata)
                    counts[idx] = counts[idx]
                    frequencies[idx] = frequencies[idx]

                    print("{0}\t{1}\t{2}\t{3}\t{4}".format(i, idx, kmer_ids[idx], counts[idx], frequencies[idx]))
                    kdb_out.write("{0}\t{1}\t{2}\t{3}\t{4}\t{5}\n".format(i, idx, kmer_ids[i], counts[idx], frequencies[idx], json.dumps(kmer_metadata)))
            else:
                for i, idx in enumerate(kmer_ids):
                    kmer_id = int(idx)
                    logger.info("{0}\t{1}\t{2}\t{3}\t{4}".format(i, idx, kmer_ids[i], counts[idx], frequencies[idx]))
                    p = profile[i]
                    c = counts[i]
                    f = frequencies[i]
                    seq = kmer.id_to_kmer(int(kmer_id), arguments.k)
                    kmer_metadata = kmer.neighbors(seq, arguments.k) # metadata is initialized by the neighbors
                    all_metadata.append(kmer_metadata)
                    print("{0}\t{1}\t{2}\t{3}\t{4}".format(i, idx, kmer_ids[i], counts[i], frequencies[i]))
                    kdb_out.write("{0}\t{1}\t{2}\t{3}\t{4}\t{5}\n".format(i, idx, kmer_ids[i], counts[i], frequencies[i], json.dumps(kmer_metadata)))
        logger.info("Wrote 4^k = {0} k-mer counts + neighbors to the .kdb file.".format(total_kmers))

        logger.info("Done")
            
    finally:
        kdb_out._write_block(kdb_out._buffer)
        kdb_out._handle.flush()
        kdb_out._handle.close()

        sys.stderr.write("Total k-mers processed: {0}\n".format(all_observed_kmers))
        sys.stderr.write("Final nullomer count:   {0}\n".format(total_nullomers))
        sys.stderr.write("Unique {0}-mer count:     {1}\n".format(arguments.k, unique_kmers))
        sys.stderr.write("Total {0}-mer count:     {1}\n".format(arguments.k, total_kmers))
    

        sys.stderr.write("\nDone\n")
    
        
def get_root_logger(level):
    levels=[logging.WARNING, logging.INFO, logging.DEBUG]
    if level < 0 or level > 2:
        raise TypeError("{0}.get_root_logger expects a verbosity between 0-2".format(__file__))
    logging.basicConfig(level=levels[level], format="%(levelname)s: %(asctime)s %(funcName)s L%(lineno)s| %(message)s", datefmt="%Y/%m/%d %I:%M:%S")
    root_logger = logging.getLogger()
    for name in logging.Logger.manager.loggerDict.keys():
        if ('boto' in name) or ('urllib3' in name) or ('s3' in name) or ('findfont' in name):
            logging.getLogger(name).setLevel(logging.WARNING)

    return root_logger


def citation_info():
    import pkg_resources
    citation = None
    if pkg_resources.resource_exists('kmerdb', 'CITATION'):
        citation = pkg_resources.resource_string('kmerdb', 'CITATION').decode('utf-8').rstrip()
        if citation == "":
            return
        else:
            sys.stderr.write("Printing citation notice to stderr. This will not interfere with the execution of the program in any way. Please see CITATION_FAQ.md for any questions.\n")
            sys.stderr.write(citation + "\n\n\n")
            sys.stderr.write("Run 'kmerdb citation' to silence.\n")
    else:
        raise IOError("Cannot locate the extra package data file 'kmerdb/CITATION', which should have been distributed with the program")


def cli():
    sys.stderr.write("Running kdb script from '{0}'\n".format(__file__))
    sys.stderr.write("Checking installed environment...\n")
    primary_path = sys.path[0]
    ver_info = sys.version_info
    py_version = "python{0}.{1}".format(ver_info.major, ver_info.minor)
    #sys.path.append(os.path.join(primary_path, "../lib/{0}/site-packages".format(py_version)))
    #sys.path.append(os.path.join(os.path.dirname(__file__), "../lib"))
    #site_packages = [p for p in sys.path if "kdb" in p]
    
    #sys.stderr.write("PYTHONPATH={0}".format(sys.path))
    #sys.path.remove(os.path.dirname(os.path.abspath(__file__)))
    citation_info()

    
    parser = argparse.ArgumentParser()

    subparsers = parser.add_subparsers(help="Use --help with sub-commands")


    profile_parser = subparsers.add_parser("profile", help="Parse data into the database from one or more sequence files")
    profile_parser.add_argument("-v", "--verbose", help="Prints warnings to the console by default", default=0, action="count")
    profile_parser.add_argument("-k", default=12, type=int, help="Choose k-mer size (Default: 12)")

    profile_parser.add_argument("-p", "--parallel", type=int, default=1, help="Shred k-mers from reads in parallel")

    profile_parser.add_argument("--batch-size", type=int, default=100000, help="Number of updates to issue per batch to PostgreSQL while counting")
    profile_parser.add_argument("-b", "--fastq-block-size", type=int, default=100000, help="Number of reads to load in memory at once for processing")
    profile_parser.add_argument("-n", type=int, default=1000, help="Number of k-mer metadata records to keep in memory at once before transactions are submitted, this is a space limitation parameter after the initial block of reads is parsed. And during on-disk database generation")
    #profile_parser.add_argument("--keep-S3-file", action="store_true", help="Download S3 file to the current working directory")
    profile_parser.add_argument("--keep-db", action="store_true", help=argparse.SUPPRESS)
    profile_parser.add_argument("--both-strands", action="store_true", default=False, help="Retain k-mers from the forward strand of the fast(a|q) file only")
    profile_parser.add_argument("--all-metadata", action="store_true", default=False, help="Include read-level k-mer metadata in the .kdb")
    profile_parser.add_argument("--sorted", action="store_true", default=False, help="Sort the output kdb file by count")
    #profile_parser.add_argument("--sparse", action="store_true", default=False, help="Whether or not to store the profile as sparse")

    profile_parser.add_argument("seqfile", nargs="+", type=str, metavar="<.fasta|.fastq>", help="Fasta or fastq files")
    profile_parser.add_argument("kdb", type=str, help="Kdb file")
    profile_parser.set_defaults(func=profile)

    header_parser = subparsers.add_parser("header", help="Print the YAML header of the .kdb file and exit")
    header_parser.add_argument("-v", "--verbose", help="Prints warnings to the console by default", default=0, action="count")
    header_parser.add_argument("-j", "--json", help="Print as JSON. DEFAULT: YAML")
    header_parser.add_argument("kdb", type=str, help="A k-mer database file (.kdb)")
    header_parser.set_defaults(func=header)
    
    view_parser = subparsers.add_parser("view", help="View the contents of the .kdb file")
    view_parser.add_argument("-v", "--verbose", help="Prints warnings to the console by default", default=0, action="count")
    view_parser.add_argument("-H", "--header", action="store_true", help="Include header in the output")
    view_parser.add_argument("--sorted", action="store_true", help="Sort the k-mer profile before displaying")
    view_parser.add_argument("--dtype", type=str, default="uint64", help="Read in the profiles as unsigned integer 64bit NumPy arrays")
    view_parser.add_argument("-d", "--decompress", action="store_true", help="Decompress input? DEFAULT: ")
    view_parser.add_argument("-c", "--compress", action="store_true", help="Print compressed output")
    view_parser.add_argument("kdb_in", type=str, nargs="?", default=None, help="A k-mer database file (.kdb) to be read (Optional)")
    view_parser.add_argument("kdb_out", type=str, nargs="?", default=None, help="A k-mer database file (.kdb) to be written to (Optional)")
    view_parser.set_defaults(func=view)


    matrix_parser = subparsers.add_parser("matrix", help="Generate a reduced-dimensionality matrix of the 4^k * n (k-mers x samples) data matrix.")
    matrix_parser.add_argument("-v", "--verbose", help="Prints warnings to the console by default", default=0, action="count")
    matrix_parser.add_argument("--with-index", default=False, action="store_true", help="Print the row indices as well")
    matrix_parser.add_argument("--column-names", default=None, type=str, help="A filepath to a plaintext flat file of column names.")
    matrix_parser.add_argument("--delimiter", default="\t", type=str, help="The choice of delimiter to parse the input .tsv with. DEFAULT: '\t'")
    matrix_parser.add_argument("-d", "--dtype", default="uint32", type=str, choices=["uint32", "uint64", "float32", "float64"], help="Data type to read into memory. DEFAULT: 'uint32'")
    matrix_parser.add_argument("--output-delimiter", type=str, default="\t", help="The output delimiter of the final csv/tsv to write. DEFAULT: '\t'")
        
    #matrix_parser.add_argument("--normalize-with", type=str, choices=["ecopy", "DESeq2"], default="DESeq2", help="Normalize with which method? DEFAULT: DESeq2")
    matrix_parser.add_argument("--no-normalized-ints", action="store_true", default=False, help="Don't round normalized counts to the nearest integer")
    matrix_parser.add_argument("-k", default=None, type=int, help="The k-dimension that the files have in common")
    matrix_parser.add_argument("-n", default=None, type=int, help="The number of dimensions to reduce with PCA or t-SNE. DEFAULT: an elbow graph will be generated if -n is not provided to help the user choose -n")

    matrix_parser.add_argument("--perplexity", default=5, type=int, help="The choice of the perplexity for t-SNE based dimensionality reduction")
    matrix_parser.add_argument("method", choices=["PCA", "tSNE", "DESeq2", "pass", "Frequency"], default=None, help="Choice of distance metric between two profiles")
    matrix_parser.add_argument("input", nargs="*", default=[], metavar="<kdbfile1 kdbfile2 ...|input.tsv|STDIN>", help="Two or more .kdb files, or another count matrix in tsv/csv")
    matrix_parser.set_defaults(func=get_matrix)
    
    # rarefy_parser = subparsers.add_parser("rarefy", help="Generate rarefaction information using ecopy.diversity.rarefy for the supplied .kdb files")
    # rarefy_parser.add_argument("-v", "--verbose", help="Prints warnings to the console by default", default=0, action="count")
    # rarefy_parser.add_argument("-d", "--delimiter", default="\t", type=str, help="The choice of delimiter to parse the DataFrame with")
    # rarefy_parser.add_argument("--output-delimiter", type=str, default="\t", help="The output delimiter of the final csv/tsv to write.")
    # rarefy_parser.add_argument("-i", "--input", type=argparse.FileType("r"), default=None, help="The input reduced dimension or simply normalized matrix to use with K-means clustering")
    # rarefy_parser.add_argument("-o", "--output", type=argparse.FileType("w"), default=None, help="THe output csv/tsv of rarefied data")
    # rarefy_parser.set_defaults(func=rarefy)

    kmeans_parser = subparsers.add_parser("kmeans", help="Cluster the files according to their k-mer profile")
    kmeans_parser.add_argument("-v", "--verbose", help="Prints warnings to the console by default", default=0, action="count")
    kmeans_parser.add_argument("-d", "--delimiter", type=str, default="\t", help="The delimiter of the input csv/tsv to parse, presumably produced by 'kdb matrix'.")
    kmeans_parser.add_argument("--output-delimiter", type=str, default="\t", help="The output delimiter of the final csv/tsv to write.")
    kmeans_parser.add_argument("--distance", type=str, default='e', choices=['e', 'b', 'c', 'a', 'u', 'x', 's', 'k'], help="Different distance metrics offered by kcluster. It is highly advised that you check both this source and their documentation to see how this is implemented.")

    
    kmeans_parser.add_argument("-k", default=None, type=int, help="The choice of k for clustering", required=True)
    kmeans_parser.add_argument("-i", "--input", type=argparse.FileType("r"), default=None, help="The input reduced dimension or mereley normalized matrix to use with K-means clustering")
    kmeans_parser.add_argument("-o", "--output", type=argparse.FileType("w"), default=None, help="The output csv/tsv with added 'label' to use for graphing in R, if the matplotlib graphs are not sufficient.")
    kmeans_parser.add_argument("method", choices=["sklearn", "Biopython"], default="Biopython", help="The Python implementation of k-means to use. The Biopython method is selected for access to alternative distance metrics")
    kmeans_parser.set_defaults(func=kmeans)

    hierarchical_parser = subparsers.add_parser("hierarchical", help="Use scipy.cluster.hierarchy to generate a dendrogram from a distance matrix")
    hierarchical_parser.add_argument("-v", "--verbose", help="Prints warnings to the console by default", default=0, action="count")
    hierarchical_parser.add_argument("-d", "--delimiter", type=str, default="\t", help="The delimiter to use when reading the csv.")
    hierarchical_parser.add_argument("-i", "--input", type=argparse.FileType("r"), default=None, help="The input distance matrix for hierarchical clustering")
    hierarchical_parser.add_argument("-m", "--method", type=str, choices=["single", "complete", "average", "weighted", "centroid", "median", "ward"], default="ward", help="The method for linkage fitting in R to use")
    hierarchical_parser.set_defaults(func=hierarchical)
    
    dist_parser = subparsers.add_parser("distance", help="Calculate various distance metrics between profiles")
    dist_parser.add_argument("-v", "--verbose", help="Prints warnings to the console by default", default=0, action="count")
    dist_parser.add_argument("--output-delimiter", type=str, default="\t", help="The output delimiter of the final csv/tsv to write.")
    dist_parser.add_argument("-d", "--dtype", type=str, default="uint32", choices=["uint32", "uint64", "float32", "float64"], help="Choice of data type to read data into memory with. Default: 'uint32'")
    dist_parser.add_argument("--column-names", type=str, default=None, help="A filepath to a plaintext flat file of column names.")
    dist_parser.add_argument("--delimiter", type=str, default="\t", help="The delimiter to use when printing the csv.")
    dist_parser.add_argument("-k", default=None, type=int, help="The k-dimension that the files have in common")
    
    dist_parser.add_argument("metric", choices=[
        "braycurtis",
        "canberra",
        "chebyshev",
        "cityblock",
        "correlation",
        "cosine",
        "dice",
        "euclidean",
        "hamming",
        "jaccard",
        "jensenshannon",
        "kulsinski",
        "mahalanobis",
        "matching",
        "minkowski",
        "pearson",
        "rogerstanimoto"
        "russelrao",
        "seuclidean",
        "sokalmichener",
        "sokalsneath",
        "spearman",
        "sqeuclidean",
        "yule"], default="correlation", help="Choice of distance metric between two profiles")
    dist_parser.add_argument("input", nargs="*", default=[], metavar="<kdbfile1 kdbfile2 ...|input.tsv|STDIN>", help="Two or more .kdb files, or another count matrix in tsv/csv")
    dist_parser.set_defaults(func=distances)


    index_parser = subparsers.add_parser("index", help="Create a index file that can be held in memory")
    index_parser.add_argument("-v", "--verbose", help="Prints warnings to the console by default", default=0, action="count")
    index_parser.add_argument("kdb", type=str, help="A k-mer database file (.kdb)")
    index_parser.set_defaults(func=index_file)

    shuf_parser = subparsers.add_parser("shuf", help="Create a shuffled .kdb file")
    shuf_parser.add_argument("-v", "--verbose", help="Prints warnings to the console by default", default=0, action="count")
    shuf_parser.add_argument("kdb_in", type=str, help="An indexed k-mer database file (.kdb)")
    shuf_parser.add_argument("kdb_out", type=str, help="The output shuffled k-mer database file (.kdb)")
    shuf_parser.set_defaults(func=shuf)

    
    markov_probability_parser = subparsers.add_parser("probability", help=u"""
Calculate the log-odds ratio of the Markov probability of a given sequence from the product (pi) of the transition probabilities(aij) times the frequency of the first k-mer (P(X1)), given the entire k-mer profile of a species.

See https://matthewralston.github.io/quickstart#kmerdb-probability for more details.

1. Durbin, R., Eddy, S.R., Krogh, A. and Mitchison, G., 1998. Biological sequence analysis: probabilistic models of proteins and nucleic acids. Cambridge university press.
""")

    markov_probability_parser.add_argument("-v", "--verbose", help="Prints warnings to the console by default", default=0, action="count")
    markov_probability_parser.add_argument("-d", "--delimiter", type=str, default="\t", help="The delimiter to use when reading the csv.")
    markov_probability_parser.add_argument("-b", "--fastq-block-size", type=int, default=100000, help="Number of reads to load in memory at once for processing")
    markov_probability_parser.add_argument("seqfile", type=str, metavar="<.fasta|.fastq>", default=None, help="Sequences to calculate standard Markov-chain probabilities from, either .fasta or .fastq")
    markov_probability_parser.add_argument("kdb", type=str, help="An indexed k-mer database file (.kdb)")
    markov_probability_parser.set_defaults(func=markov_probability)

    citation_parser = subparsers.add_parser("citation", help="Silence the citation notice on further runs")
    citation_parser.add_argument("-v", "--verbose", help="Prints warnings to the console by default", default=0, action="count")
    citation_parser.set_defaults(func=citation)
    
    args=parser.parse_args()
    global logger
    logger = get_root_logger(args.verbose)

    sys.stderr.write("Constructed a logger for the program...\n")
    #logger.debug(sys.path)
    args.func(args)
    



'''
   Copyright 2020 Matthew Ralston

   Licensed under the Apache License, Version 2.0 (the "License");
   you may not use this file except in compliance with the License.
   You may obtain a copy of the License at

       http://www.apache.org/licenses/LICENSE-2.0

   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   See the License for the specific language governing permissions and
   limitations under the License.

'''



import logging
logger = logging.getLogger(__file__)


import io
import os
import sys
import gzip
import hashlib

from Bio import SeqIO
import Bio

class SeqParser:
    """
    Largely independent module, needs 3 pieces of information passed back in from the outside
    """
    def __init__(self, filepath, num, k):
        if type(filepath) is not str:
            raise TypeError("kdb.seqrecord.SeqParser expects a str as its first positional argument")
        elif type(num) is not int:
            raise TypeError("kmerdb.seqparser.SeqParser expects an int as its second positional argument")
        elif type(k) is not int:
            raise TypeError("kmerdb.seqparser.SeqParser expects an int as its third positional argument")
        
        self.k = k
        self.num = num
        self.reads = []
        # Header items
        self.filepath = filepath
        self.md5 = None
        self.sha256 = None
        self.total_reads = 0
        self.total_kmers = 0
        self.unique_kmers = 0
        self.nullomers = 0
        self.mononucleotides = {'A': 0, 'C': 0, 'G': 0, 'T': 0}
        # 
        self.compressed = False
        self.fastq = False
        exts = os.path.splitext(filepath)

        if exts[-1] == ".gz":
            self.compressed = True
            nogzexts = os.path.splitext(exts[0])
            if nogzexts[-1] == ".fq" or nogzexts[-1] == ".fastq":
                self.fastq = True
            elif nogzexts[-1] == ".fna" or nogzexts[-1] == ".fasta" or exts[-1] == ".fa":
                self.fastq = False
            else:
                raise ValueError("Cannot parse files of extension '{0}'.\n\nRequires fasta (.fna, .fasta, .fa), fastq (.fq, .fastq), or their gzipped equivalents")
        else: # Must be fasta or fastq uncompressed
            if exts[-1] == ".fq" or exts[-1] == ".fastq":
                self.fastq = True
            elif exts[-1] == ".fna" or exts[-1] == ".fasta" or exts[-1] == ".fa":
                self.fastq = False
            else:
                raise ValueError("Cannot parse files of extension '{0}'.\n\nRequires fasta (.fna, .fasta, .fa), fastq (.fq, .fastq), or their gzipped equivalents")
        if self.fastq:
            self.__class__.__iter__ = self._iter_fastq
            self.__class__.__next__ = self._next_fastq
        else:
            self.__class__.__iter__ = self._iter_fasta
            self.__class__.__next__ = self._next_fasta

                

                
        if self.compressed:
            if self.fastq:
                logger.info("Opening gzipped fastq file '{0}'...".format(filepath))
                self._handle = SeqIO.parse(gzip.open(self.filepath, 'rt'), "fastq")
            else:
                logger.info("Opening gzipped fasta file '{0}'...".format(filepath))
                self._handle = SeqIO.parse(gzip.open(self.filepath, 'rt'), "fasta")
        else:
            if self.fastq:
                logger.info("Opening uncompressed fastq file '{0}'...".format(filepath))
                self._handle = SeqIO.parse(open(self.filepath, 'r'), "fastq")
            else:
                logger.info("Opening uncompressed fasta file '{0}'...".format(filepath))
                self._handle = SeqIO.parse(open(self.filepath, 'r'), "fasta")

        # Get checksums
        self.md5, self.sha256 = self.__checksum()


    def __checksum(self):
        """Generates md5 and sha256 checksums of a file
        :returns: (md5, sha256)
        :rtype: tuple
        """
        if not os.path.exists(self.filepath):
            raise IOError("kmerdb.seqparser.FastqParser.__checksum could not find '{}' on the filesystem".format(filepath))
        hash_md5 = hashlib.md5()
        hash_sha256 = hashlib.sha256()
        with open(self.filepath, 'rb') as ifile:
            for chunk in iter(lambda: ifile.read(4096), b""):
                hash_md5.update(chunk)
                hash_sha256.update(chunk)
        return (hash_md5.hexdigest(), hash_sha256.hexdigest())

    def header_dict(self):
        """ Create a header dictionary to convert into YAML to go in the header block of the compression header. Has a schema to be validated, defined in config.py

        :returns: dict
        :rtype: 

        """
        return {
            "filename": self.filepath,
            "md5": self.md5,
            "sha256": self.sha256,
            "total_reads": self.total_reads,
            "total_kmers": self.total_kmers,
            "unique_kmers": self.unique_kmers or 4**self.k - self.nullomers,
            "nullomers": self.nullomers,
            "mononucleotides": self.mononucleotides
        }
            
    def __exit__(self, exc_type, exc_value, traceback):
        self._handle.close()

    def __enter__(self):
        return self
        
        
    def _iter_fastq(self):
        try:
            for i in range(self.num):
                self.reads.append(self._handle.__next__())
        except StopIteration as e:
            # with db._engine.connect() as conn:
            #     self.nullomers = db.get_nullomers(conn).pop()
            #     self.total_kmers = db.get_sum_counts(conn).pop()
            #     self.unique_kmers = 4**k - self.nullomers
            pass
        except ValueError as e:
            logger.error(e)
            logger.error("\n\nFastq format error: '{0}' seems to not be fastq format\n\n")
            raise e
        return self

    def _next_fastq(self):
        if not len(self.reads):
            raise StopIteration
        else:
            self.total_reads += 1
            #sys.stderr.write("\r")
            #sys.stderr.write("Read {0} reads from '{1}'...".format(self.total_reads, self.filepath))
            read = self.reads.pop()
            self.mononucleotides["A"] += read.seq.count('A')
            self.mononucleotides["T"] += read.seq.count('T')
            self.mononucleotides["C"] += read.seq.count('C')
            self.mononucleotides["G"] += read.seq.count('G')
            # If you needed to see the fastq mononucleotides go up during parsing
            #sys.stderr.write("\r")
            #sys.stderr.write("Mononucleotides | A: {0}, C: {1}, G: {2}, T{3}".format(self.mononucleotides["A"], self.mononucleotides["C"], self.mononucleotides["G"], self.mononucleotides["T"]))
            return read

    def _iter_fasta(self):
        return self

    def _next_fasta(self):

        seq = next(self._handle)
        self.mononucleotides["A"] += seq.seq.count('A')
        self.mononucleotides["T"] += seq.seq.count('T')
        self.mononucleotides["C"] += seq.seq.count('C')
        self.mononucleotides["G"] += seq.seq.count('G')

        logger.debug("Mononucleotides | A: {0}, C: {1}, G: {2}, T: {3}".format(self.mononucleotides["A"], self.mononucleotides["C"], self.mononucleotides["G"], self.mononucleotides["T"]))
        self.total_reads += 1
        sys.stderr.write("\r")
        sys.stderr.write("Read {0} sequences from '{1}'...\n".format(self.total_reads, self.filepath))
        
        return seq
        
        




'''
   Copyright 2020 Matthew Ralston

   Licensed under the Apache License, Version 2.0 (the "License");
   you may not use this file except in compliance with the License.
   You may obtain a copy of the License at

       http://www.apache.org/licenses/LICENSE-2.0

   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   See the License for the specific language governing permissions and
   limitations under the License.

'''


import os
import sys
import yaml
import json
import time
from datetime import datetime
from math import ceil
from itertools import chain, repeat


# from sqlalchemy.orm import sessionmaker
# from sqlalchemy.orm.attributes import flag_modified
# from sqlalchemy.ext.declarative import declarative_base
# from sqlalchemy import Table, Column, Integer, String, MetaData, ForeignKey, Sequence, JSON, Boolean

#from psycopg2 import sql
import tempfile
import numpy as np


from kmerdb import seqparser, kmer

import logging
logger = logging.getLogger(__file__)





def parsefile(filepath:str, k:int, rows_per_batch:int=100000, b:int=50000, n:int=1000, both_strands:bool=False, all_metadata:bool=False):
    """Parse a single sequence file in blocks/chunks with multiprocessing support

    :param filepath: Path to a fasta or fastq file
    :type filepath: str
    :param k: Choice of k to shred k-mers with
    :type k: int
    :param b: Number of reads (per block) to process in parallel
    :type b: int
    :param stranded: Strand specificity argument for k-mer shredding process
    :type stranded: bool
    :returns: (counts, header_dictionary, nullomers) header_dictionary is the file's metadata for the header block
    :rtype: (numpy.ndarray, dict, list)

    """

    from kmerdb import kmer
    if type(filepath) is not str:
        raise TypeError("kmerdb.parse.parsefile expects a str as its first positional argument")
    elif not os.path.exists(filepath):
        raise OSError("kmerdb.parse.parsefile could not find the file '{0}' on the filesystem".format(filepath))
    elif type(k) is not int:
        raise TypeError("kmerdb.parse.parsefile expects an int as its second positional argument")
    elif type(b) is not int:
        raise TypeError("kmerdb.parse.parsefile expects the keyword argument 'b' to be an int")
    elif type(n) is not int:
        raise TypeError("kmerdb.parse.parsefile expects the keyword argument 'n' to be an int")
    elif type(both_strands) is not bool:
        raise TypeError("kmerdb.parse.parsefile expects the keyword argument 'both_strands' to be a bool")
    elif type(all_metadata) is not bool:
        raise TypeError("kmerdb.parse.parsefile expects the keyword argument 'all_metadata' to be a bool")
    data = {} # This is the dictionary of tuples, keyed on k-mer id, and containing 3-tuples ('kmer_id', 'read/start/reverse')
    keys = set()
    rows = []
    total_kmers = 4**k
    nullomers = set()
    # Build fasta/fastq parser object to stream reads into memory
    logger.debug("Initializing parser...")
    seqprsr = seqparser.SeqParser(filepath, b, k)
    fasta = not seqprsr.fastq # Look inside the seqprsr object for the type of file

    # Initialize the kmer array
    try:
        counts = np.zeros(total_kmers, dtype="uint64")
    except TypeError as e:
        logger.error("Invalid dtype for numpy array instantiation")
        logger.error(e)
        raise e

    logger.info("Successfully allocated space for {0} unsigned integers: {1} bytes".format(total_kmers, counts.nbytes))
        # Instantiate the kmer class
    Kmer = kmer.Kmers(k, strand_specific=not both_strands, fasta=fasta, all_metadata=all_metadata) # A wrapper class to shred k-mers with

    recs = [r for r in seqprsr] # A block of exactly 'b' reads-per-block to process in parallel
    if not fasta:
        logger.debug("Read exactly {0} records from the seqparser object".format(len(recs)))
        assert len(recs) <= b, "The seqparser should return exactly {0} records at a time".format(b)
    else:
        logger.debug("Skipping the block size assertion for fasta files")
    logger.info("Read {0} sequences from the {1} seqparser object".format(len(recs), "fasta" if fasta else "fastq"))
    all_kmer_metadata = list([] for x in range(total_kmers))
    while len(recs): # While the seqprsr continues to produce blocks of reads

        num_recs = len(recs)
        logger.info("Processing a block of {0} reads/sequences".format(num_recs))

        # Initialize an all_kmer_metadata list
        kmer_metadata = []
        list_of_dicts = list(map(Kmer.shred, recs))

        # Flatmap to 'kmers', the dictionary of {'id': read_id, 'kmers': [ ... ]}
        logger.debug("\n\nAcquiring linked-list of all k-mer ids from {0} sequence records...\n\n".format(num_recs))
        kmer_ids = list(chain.from_iterable(map(lambda x: x['kmers'], list_of_dicts)))
        logger.debug("Successfully allocated linked-list for all k-mer ids read.")
        sus = set()
        logger.info("Removing any eroneous k-mers without an id, cause by 'N' content, creating new gaps in the k-mer profile...")
        logger.info("Will substitute IUPAC residues with their respective residues, adding one count for each")
        num_sus = 0
        for i, k in enumerate(kmer_ids):
            if k is None:
                sus.add(i) # This propagates the N removal, by adding those to the set for removal later.
                num_sus += 1

        logger.info("Created {0} gaps in the k-mer profile to clean it of N-content".format(num_sus))
        if num_sus > 0:
            logger.warning("{0} uncountable k-mer identified".format(num_sus))
            logger.warning("Likely caused by the presence of the unspecified nucleotide 'N' in a .fasta file")
            logger.warning("Will remove from the final committed profile")

        ########################
        # K-mer cleaning
        #
        # 
        # The following procedure, removes k-mers that don't have a formal k-mer id
        # More specifically, it eliminates entries that were formally loaded into the k-mers array,
        # and thus would be a minor percentage of the total number of k-mers,
        # Specifically, k-mers or subsequences with N can be excluded from the profile
        # To essentially eliminate an area of sequence and create an artificial gap in the "retained" graph.
        # The graph can be explicitly collapsed when you want to, by supplying reads or sequence information, that is contiguous "in your judgement".
        # By introducing these artificial gaps, we expose the errors or unspecified bases through data.
        #
        logger.info("Found {0} invalid or enumerable k-mers, sequences which have no defined index, i".format(num_sus))
        logger.debug("In other words, one or more characters in your .fasta or .fastq file were 'N' or otherwise contained IUPAC characters that need to be substituted.")
        logger.debug("We have chosen to omit k-mer with unspecified nucleotides 'N', and these make gaps in our database explicitly.")
        logger.info("These can be recovered from the reads if they are needed, and the read ids may be found with the experimental --all-metadata flag")
        if all_metadata:

            logger.warning("\n\n\nGENERATING ALL METADATA\n\n\nThis is extremely expensive and experimental. You have been warned.\n\n\n")
            reads = list(chain.from_iterable(map(lambda x: x['seqids'], list_of_dicts)))
            starts = list(chain.from_iterable(map(lambda x: x['starts'], list_of_dicts)))
            reverses = list(chain.from_iterable(map(lambda x: x['reverses'], list_of_dicts)))
            logger.debug("Checking each k-mer for N-content and IUPAC substitution")
            for i, x in enumerate(kmer_ids): # This removes N content
                if i in sus: # This is where we actually delete the N content, in case that is eventually supported.
                    kmer_ids[i] = None
                    reads[i] = None
                    starts[i] = None
                    reverses[i] = None
            kmer_ids = list(filter(lambda k: k is not None, kmer_ids)) 
            reads = list(filter(lambda r: r is not None, reads))
            starts = list(filter(lambda s: s is not None, starts))
            reverses = list(filter(lambda r: r is not None, reverses))
        else:
            logger.debug("Checking each k-mer for IUPAC substitution or N content.")
            for i, x in enumerate(kmer_ids): # Here we remove the k-mer ids where N-content is detected, in case they are needed, you can use kmer_ids prior to this point to build functionality.
                if i in sus:
                    kmer_ids[i] = None
            logger.info("Eliminating suspicious 'sus' k-mers, i.e. those with N-content or IUPAC substitutions")
            kmer_ids = list(filter(lambda k: k is not None, kmer_ids))
            reads = [] # I'm keeping this in, just in case for some reason the variable names are needed in the 
            starts = []
            reverses = []
            if None in kmer_ids:
                logger.debug("In the no-metadata field")
                # Actually they were just introduced to be filtered out, instead of deleted
                # Because each deletion whould cange the array index
                # So instead we set ghtme to None, and filter out
                raise ValueError("K-mer ids should never actually be none.")

        #logger.debug(kmer_ids)
        logger.debug("{0} 'clean' kmers were identified successfully from {1} input sequences".format(len(kmer_ids), num_recs))
        logger.debug("Flatmapped {0} kmers for their metadata aggregation".format(len(kmer_ids), len(starts)))
        # Assert that all list lengths are equal before adding metadata to k-mers
        if all_metadata is True and len(kmer_ids) == len(reads) and len(reads) == len(starts) and len(starts) == len(reverses):
            N = len(starts)
                
            kmer_metadata = list(zip(kmer_ids, reads, starts, reverses))
            # Everything is in the right order
            logger.warning("Dumping all metadata into the .kdb file eventually. This could be expensive...")

        elif not all_metadata and len(reads) == 0 and len(starts) == 0 and len(reverses) == 0:
            logger.debug("Skipping metadata allocation")
            pass # If we're not doing metadata, don't do it
        else: # Raise an error if the numbers of items per list are not equal
            logger.error("{0} kmer ids".format(len(kmer_ids)))
            logger.error("{0} sequence/read associations".format(len(reads)))
            logger.error("{0} start positions for the associations found".format(len(starts)))
            logger.error("{0} reverse bools for each association".format(len(reverses)))
                
            raise ValueError("Unexpectedly, the number of ids did not match up with the number of other metadata elements per k-mer OR other unknown error")


        # else:
        #     raise RuntimeError("Still have no clue what's going on...")
        # On disk k-mer counting
        # Thank you, this was brilliant
        # https://stackoverflow.com/a/9294062/12855110
        # 01-01-2022 This has been removed in favor of in-memory counting
        num_kmers = len(kmer_ids)

        if num_kmers == 0:
            raise ValueError("No k-mers to add. Something likely went wrong. Please report to the issue tracker")
        else:
            sys.stderr.write("\nAccumulating all k-mers from this set of records...")
            for kmer in kmer_ids:
                counts[kmer] += 1
        # all_kmer_metadata
        if all_metadata:
            for single_kmer_id, read, start, reverse in kmer_metadata:
                all_kmer_metadata[single_kmer_id].append((read, start, reverse))


        recs = [r for r in seqprsr] # The next block of exactly 'b' reads
        # This will be logged redundantly with the sys.stderr.write method calls at line 141 and 166 of seqparser.py (in the _next_fasta() and _next_fastq() methods)
        #sys.stderr("\n")
        logger.info("Read {0} more records from the {1} seqparser object".format(len(recs), "fasta" if fasta else "fastq"))
    # Get nullomers
    # only nullomer counts
    unique_kmers = int(np.count_nonzero(counts))
    num_nullomers = total_kmers - unique_kmers
    is_nullomer = np.where(counts == 0)
    nullomers = []
    for i, j in enumerate(is_nullomer):
        if j is True:
            nullomers.append(i)
    seqprsr.total_kmers = int(np.sum(counts))
    seqprsr.unique_kmers = unique_kmers
    seqprsr.nullomers = num_nullomers
    sys.stderr.write("\n\n\nFinished counting k-mers{0} from '{1}'...\n\n\n".format(' and metadata' if all_metadata else '', filepath))

    return counts, seqprsr.header_dict(), nullomers, all_kmer_metadata




class Parseable:
    def __init__(self, arguments):
        self.arguments = arguments
            
        
    def parsefile(self, filename):
        """Wrapper function for parse.parsefile to keep arguments succinct for deployment through multiprocessing.Pool
            
        :param filename: the filepath of the fasta(.gz)/fastq(.gz) to process with parsefile -> seqparser
        :type filename: str
        :returns: (db, m, n)
        """
        return parsefile(filename, self.arguments.k, rows_per_batch=self.arguments.batch_size, b=self.arguments.fastq_block_size, n=self.arguments.n, both_strands=self.arguments.both_strands, all_metadata=self.arguments.all_metadata)





'''
   Copyright 2020 Matthew Ralston

   Licensed under the Apache License, Version 2.0 (the "License");
   you may not use this file except in compliance with the License.
   You may obtain a copy of the License at

       http://www.apache.org/licenses/LICENSE-2.0

   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   See the License for the specific language governing permissions and
   limitations under the License.

'''

import sys
from itertools import chain, repeat
import logging
logger = logging.getLogger(__file__)
from Bio import SeqIO, Seq
import Bio



#############################
#
# Dictionaries/maps
#
#############################

letterToBinary={ # Unicode UTF-8 byte codes for ACGT
    65: 0,
    67: 1,
    71: 2,
    84: 3
}
binaryToLetter=['A', 'C', 'G', 'T']
standard_letters=set("ACTG")


IUPAC_TO_DOUBLET = {
    "R": ["A", "G"],
    "Y": ["C", "T"],
    "S": ["G", "C"],
    "W": ["A", "T"],
    "K": ["G", "T"],
    "M": ["A", "C"]
}
IUPAC_DOUBLET_CHARS=IUPAC_TO_DOUBLET.keys()
IUPAC_TO_TRIPLET = {
    "B": ["C", "G", "T"],
    "D": ["A", "G", "T"],
    "H": ["A", "C", "T"],
    "V": ["A", "C", "G"]
}
IUPAC_TRIPLET_CHARS=IUPAC_TO_TRIPLET.keys()
n = "N"
#############################
#
# Lambdas
#
#############################


get_doublet_chars = lambda x: IUPAC_TO_DOUBLET[x] # list
get_triplet_chars = lambda x: IUPAC_TO_TRIPLET[x]
replace_char = lambda seq, x, y: seq.replace(x, y) # str

#############################
#
# Class definition
#
#############################

class Kmers:
    """A wrapper class to pass variables through the multiprocessing pool
    
    :ivar k: The choice of k to shred with
    :ivar strand_specific: Include k-mers from forward strand only
    """
    def __init__(self, k, strand_specific=True, fasta=False, all_metadata=False):
        """

        :param k: The choice of k to shred with
        :type k: int
        :param strand_specific: Include k-mers from forward strand only
        :type strand_specific: bool
        :param fasta: Whether or not to print extra logging in the case of fasta and not in the case of fastq
        :type fasta: bool
        :param all_metadata: Whether or not to pass back extra metadata
        :type all_metadata: bool


        """
        if type(k) is not int:
            raise TypeError("kmerdb.kmer.Kmers.__init__() expects an int as its first positional argument")
        elif type(strand_specific) is not bool:
            raise TypeError("kmerdb.kmer.Kmers.__init__() expects a bool as its second positional argument")
        elif type(fasta) is not bool:
            raise TypeError("kmerdb.kmer.Kmers.__init__() expects a bool as its third positional argument")
        elif type(all_metadata) is not bool:
            raise TypeError("kmerdb.kmer.Kmers.__init__() expects a bool as its fourth positional argument")
        self.k = k 
        self.strand_specific = strand_specific
        self.fasta = fasta
        self.all_metadata = all_metadata
        self.__permitted_chars = set("ACTGRYSWKMBDHVN")

    def shred(self, seqRecord):
        """

        :param seqRecord: 
        :type seqRecord: Bio.SeqRecord.SeqRecord
        :returns: 
        :rtype: 

        """
        
        
        if not isinstance(seqRecord, Bio.SeqRecord.SeqRecord):
            raise TypeError("kmerdb.kmer.Kmers expects a Bio.SeqRecord.SeqRecord object as its first positional argument")
        letters = set(seqRecord.seq)
        seqlen = len(seqRecord.seq)
        if seqlen < self.k:
            logger.error("Offending sequence ID: {0}".format(seqRecord.id))
            raise ValueError("kmerdb expects that each input sequence is longer than k.")
        all_iupac_symbols = list(letters.intersection(self.__permitted_chars) - standard_letters)
        all_non_iupac_symbols = list(letters - self.__permitted_chars)
        if len(all_non_iupac_symbols) > 0:
            logger.warning("One or more unexpected characters in the {0} sequence".format("fasta" if self.fasta else "fastq"))
            logger.warning(list(letters - self.__permitted_chars))
            raise ValueError("Non-IUPAC symbols detected in the fasta file")
        elif len(all_iupac_symbols) > 0:
            logger.warning("Will completely refuse to include k-mers with 'N'")
            logger.warning("All counts for k-mers including N will be discarded")
            logger.warning("Other IUPAC symbols will be replaced with their respective pairs/triads")
            logger.warning("And a count will be given to each, rather than a half count")
        kmers = []
        starts = []
        reverses = []
        # Each of the n-k+1 string slices become the k-mers
        for i in range(seqlen - self.k + 1):
            s = seqRecord.seq[i:(i+self.k)]
            #logger.debug(letters - self.__permitted_chars)
            iupac_symbols = list(set(s).intersection(self.__permitted_chars) - standard_letters)
            non_iupac_symbols = list(set(s) - self.__permitted_chars)

            ######################################################################33
            #       I U P A C        m o d u l e
            ######################################################################33

            if len(set(str(s)) - self.__permitted_chars) == 0 and len(iupac_symbols) > 0 and len(non_iupac_symbols) == 0:
                seqs = None
                for c in list(set(s) - standard_letters):
                    logger.debug("character: {0}".format(c))
                    logger.debug("K-mer: {0}".format(s))
                    logger.debug("IUPAC symbols in the whole sequence: {0}".format(iupac_symbols))
                    logger.debug("Non-IUPAC symbols in the whole sequence: {0}".format(non_iupac_symbols))
                    if c in IUPAC_DOUBLET_CHARS:
                        #logger.info("Doublets being replaced")
                        if seqs is None:
                            seqs = list(map(lambda x: replace_char(str(s), c, x), get_doublet_chars(c)))
                        else:
                            seqs = list(chain.from_iterable(map(lambda s: list(map(lambda x: replace_char(s, c, x), get_doublet_chars(c))), seqs)))
                    elif c in IUPAC_TRIPLET_CHARS:
                        #logger.info("Triplets being replaced")
                        if seqs is None:
                            seqs = list(map(lambda x: replace_char(str(s), c, x), get_triplet_chars(c)))
                        else:
                            seqs = list(chain.from_iterable(map(lambda s: list(map(lambda x: replace_char(s, c, x), get_triplet_chars(c))), seqs)))
                    elif c in standard_letters:
                        raise RuntimeError("Standard DNA residue detected in IUPAC module")
                    elif c == n:
                        logger.warning("N content detected")
                        continue
                    else:
                        logger.error(str(seqRecord))
                        logger.error("Full sequence above")
                        logger.error("K-mer: {0}".format(s))
                        logger.error(s)
                        raise RuntimeError("Non-IUPAC character '{0}' made it into the IUPAC extension of kmerdb.kmer.Kmer.shred".format(c))
                if seqs is None and "N" not in s:
                    logger.debug(iupac_symbols)
                    logger.debug(non_iupac_symbols)
                    logger.error(s)
                    logger.error("The k-mer above did not have any IUPAC letters")
                    raise RuntimeError("No sequences generated in IUPAC module of kmerdb.kmer.Kmer.shred")
                elif seqs is None and "N" in s:
                    continue
                one_word = "".join(seqs)
                if len(set(one_word) - self.__permitted_chars) > 0:
                    logger.error(one_word)
                    logger.error(seqs)
                    raise ValueError("Still IUPAC characters in the strings")
                else:
                    for x in seqs:
                        logger.debug(x)
                        kmers.append(kmer_to_id(x))
                        if self.all_metadata:
                            reverses.append(False)
                            starts.append(i)
                        if not self.strand_specific: # Reverse complement by default
                            kmers.append(kmer_to_id(Seq.Seq(x).reverse_complement()))
                            if self.all_metadata:
                                reverses.append(True)
                                starts.append(i)

                    # OKAY you are supposed to remember that
                    # the number of k-mer ids will no-longer match the appended metadata,
                    # So even though the lambda is correct, you have to do separate metadata for each id
            elif len(non_iupac_symbols) > 0:
                logger.debug("TEST CONDITIONS")
                logger.debug("Non-permitted characters should be 0: {0}".format(len(set(str(s)) - self.__permitted_chars)))
                logger.debug("IUPAC symbols should be >= 0: {0}".format(len(iupac_symbols)))
                logger.debug("Non-IUPAC characters should be 0: {0}".format(len(non_iupac_symbols)))
                logger.debug("Non-IUPAC symbols: {0}".format(non_iupac_symbols))
                logger.debug("IUPAC symbols: {0}".format(iupac_symbols))
                logger.debug("K-mer: {0}".format(str(s)))
                raise ValueError("Non-IUPAC symbol detected")
            else:
                try:
                    kmers.append(kmer_to_id(s))
                    if self.all_metadata:
                        reverses.append(False)
                        starts.append(i)
                    if not self.strand_specific: # Reverse complement by default
                        kmers.append(kmer_to_id(s.reverse_complement()))
                        if self.all_metadata:
                            reverses.append(True)
                            starts.append(i)
                except KeyError as e:
                    logger.debug("One or more non-IUPAC letters found their way into a part of the code they're not supposed to go")
                    logger.debug("We officially support IUPAC but the statement challenging the sequence content failed, causing a genuine runtime error")
                    logger.debug("This caused the following KeyError")
                    logger.debug(e)
                    logger.error("Letters in the sequence: {0}".format(letters))
                    logger.errror("Permitted letters: {0}".format(self.__permitted_chars))
                    logger.error("Unexpected behavior, rerun with -vv (DEBUG verbosity) to see more information")
                    raise RuntimeError("IUPAC standard extra base pairs (R, B, etc.) or non-IUPAC characters detected in the sequence")
            del s
        if self.fasta:
            sys.stderr.write("            --- ~~~ --- ~~~  shredded ~~~ --- ~~~ ---\n")
            sys.stderr.write("a {0}bp long sequence was shredded into L-k+1 {1} total and {2} unique k-mers\n\n{3} were discarded due to sequence content\n\n".format(len(seqRecord.seq), len(str(seqRecord.seq))-self.k+1, len(list(set(kmers))), len([x for x in kmers if x is None])))
        return {'id': seqRecord.id, 'kmers': kmers, "seqids": repeat(seqRecord.id, len(starts)), "starts": starts, 'reverses': reverses}



#############################
#
# Functions
#
#############################

    

def kmer_to_id(s):
    """Convert a fixed length k-mer string to the binary encoding parameterized upon that same k

    Note that the conversion of a k-mer string to an id integer
    is consistent regardless of k, 
    because the k is implicit in the k-mer string's size.

    Therefore, this method does not need to be wrapped in the k-mer class

    :param s: The input k-mer
    :type s: str
    :returns: The kPal-inspired binary encoding
    :rtype: int

    """

    if not isinstance(s, str) and type(s) is not str and not isinstance(s, Bio.Seq.Seq) and not isinstance(s, Bio.SeqRecord.SeqRecord): # Typecheck the input k-mer
        logger.error("Seq: {0}, type: {1}".format(s, type(s)))
        raise TypeError("kmerdb.kmer.kmer_to_id expects a str as its argument")
    elif s.find('N') != -1: # k-mer with 'N' do not have a binary encoding
        #logger.debug(TypeError("kdb.kmer.kmer_to_id expects the letters to contain only nucleotide symbols ATCG"))
        return None
    else: 
        idx = 0
        if isinstance(s, Bio.Seq.Seq) or isinstance(s, Bio.SeqRecord.SeqRecord):
            s = str(s)
        for c in bytes(s, "UTF-8"): # Use byteshifting for fast conversion to binary encoding
            idx = idx << 2
            try:
                idx = idx | letterToBinary[c]
            except KeyError as e:
                logger.error("Entire sequence: {0}".format(s))
                logger.error("Problematic character: {0}".format(c))
                raise e
        return idx


def id_to_kmer(id, k):
    if type(id) is not int:
        logger.error(type(id))
        raise TypeError("kmerdb.id_to_kmer expects an int as its first positional argument")
    elif type(k) is not int:
        logger.error(type(k))
        raise TypeError("kmerdb.id_to_kmer expects an int as its second positional argument")
    else:
        kmer = ""
        for i in range(k):
            kmer += binaryToLetter[id & 0x03]
            id = id >> 2
        kmer = list(kmer)
        kmer.reverse()
        return ''.join(kmer)


def neighbors(s, k):
    if not isinstance(s, str):
        raise TypeError("kmerdb.kmer.neighbors expects a Biopython Seq object as its first positional argument")
    elif type(k) is not int:
        raise TypeError("kmerdb.kmer.neighbors expects an int as its second positional argument")
    elif len(s) != k:
        raise TypeError("kmerdb.kmer.neighbors cannot calculate the {0}-mer neighbors of a {1}-mer".format(k, len(s)))
    else:
        import copy
        letters1 = copy.deepcopy(binaryToLetter)
        letters2 = copy.deepcopy(letters1)
    
        firstChar = s[0]
        lastChar  = s[-1]
        suffix    = s[1:]
        prefix    = s[:-1]
        rightNeighborIds = dict((c, kmer_to_id(suffix+c)) for c in letters1)
        leftNeighborIds = dict((c, kmer_to_id(c+prefix)) for c in letters2)
        return {"suffixes": rightNeighborIds, "prefixes": leftNeighborIds}



'''
   Copyright 2020 Matthew Ralston

   Licensed under the Apache License, Version 2.0 (the "License");
   you may not use this file except in compliance with the License.
   You may obtain a copy of the License at

       http://www.apache.org/licenses/LICENSE-2.0

   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   See the License for the specific language governing permissions and
   limitations under the License.

'''




import logging
logger = logging.getLogger(__file__)

import math
import numpy as np
import sys
#from numba import jit
import functools


from kmerdb import kmer, fileutil

identity = {
    'correlation': '1.0',
    'euclidean'  : '0.0',
    'hamming'    : '1.0',
    'pearson'    : '1.0',
    'spearman'   : '1.0'
}

def correlation(prof1:fileutil.KDBReader, prof2:fileutil.KDBReader, k=None, dtype="uint32"):
    
    if type(prof1) is not fileutil.KDBReader:
        raise TypeError("kmerdb.distance.correlation expects a KDBReader as its first positional argument")
    elif type(prof2) is not fileutil.KDBReader:
        raise TypeError("kmerdb.distance.correlation expects a KDBReader as its second positional argument")
    elif prof1.dtype != dtype:
        raise TypeError("kmerdb.distance.correlation expects a KDBReader of '{0}' as its first argument".format(dtype))
    elif prof2.dtype != dtype:
        raise TypeError("kmerdb.distance.correlation expects a KDBReader of '{0}' as its second argument".format(dtype))
    elif k is None:
        raise ValueError("Cannot calculate correlation without k")
    p1_size = int(prof1.profile.size)
    p2_size = int(prof2.profile.size)
    if p1_size == p2_size:
        #print(p1_size)
        #print(p2_size)
        #print("Trying to calculate distance. Sizes are equal.")
        #sys.exit(1)
        pass
    if k is not None and p1_size != p2_size:
        logger.error(p1_size)
        logger.error(p2_size)
        sys.exit(1)
        raise ValueError("Cannot calculate correlation of unequal size NumPy arrays.")

    N = 4**k
    if p1_size == p2_size:
        x_bar = np.sum(prof1.profile)/prof1.metadata["total_kmers"]
        y_bar = np.sum(prof2.profile)/prof2.metadata["total_kmers"]

        ssxx = 0
        ssyy = 0
        ssxy = 0
        for kmer_id in range(N):
            x = prof1.profile[kmer_id]
            y = prof2.profile[kmer_id]
            ssxx += np.square(x - x_bar)
            ssyy += np.square(y - y_bar)
            x_minus_xbar = x - x_bar
            y_minus_ybar = y - y_bar
            ssxy += x_minus_xbar*y_minus_ybar
            logger.warning("Beware custom correlations")
        #x_bar = functools.reduce(lambda a,b: a+b, prof1))
        if ssxx*ssyy == 0:
            logger.error("Incorrect denominator found, skipping...")
            return 0
        else:
            logger.info("Acquired")
            logger.debug("Well done, cap...")
            return ssxy/np.sqrt(ssxx*ssyy)
    else:
        logger.error("Cannot manage to compare a {0}-dim profile with a {1}-dim profile...".format(p1_size, p2_size))
        raise ValueError("Cannot calculate the correlation coefficient of unequal arrays.")

    
def _correlation(fname1, fname2):
    if type(fname1) is not str:
        raise TypeError("kmerdb.distance.correlation expects a str as its first positional argument")
    elif type(fname2) is not str:
        raise TypeError("kmerdb.distance.correlation expects a str as its second positional argument")
    k = None
    logger.info("Custom correlation coefficient calculation")
    with fileutil.open(fname1, mode='r') as kdb1:
        with fileutil.open(fname2, mode='r') as kdb2:
            if k is None:
                k = kdb1.metadata['k']
            if k != kdb1.metadata['k']:
                raise Exception("File '{0}' reported k = {1} instead of k = {2}".format(f, kdb1.metadata['k'], k))
            elif k != kdb2.metadata['k']:
                raise Exception("File '{0}' reported k = {1} instead of k = {2}".format(f, kdb2.metadata['k'], k))
            N = 4 ** k
            x_bar = functools.reduce(lambda a,b: a+b, map(lambda x: x['total_kmers'], kdb1.metadata['files']), 0) / N
            y_bar = functools.reduce(lambda a,b: a+b, map(lambda y: y['total_kmers'], kdb2.metadata['files']), 0) / N
            ## CALCULATE CORRELATION
            ssxx = 0
            ssyy = 0
            ssxy = 0
            for kmer_id in range(N):
                line1 = next(kdb1)
                line2 = next(kdb2)
                _, x = (int(_x) for _x in line1.rstrip().split("\t")[0:2])
                _, y = (int(_y) for _y in line2.rstrip().split("\t")[0:2])
                ssxx += np.square(x - x_bar)
                ssyy += np.square(y - y_bar)
                ssxy += (x - x_bar)*(y - y_bar)
                #logger.debug("int through np.type on custom correlation function")
                #logger.debug("Style points")
                logger.info("Squared differences.")
                logger.info("Symmetry")
            logger.debug("Sum of squared xy errors: {0}".format(ssxy))
            logger.debug("Sum of squared xx errors: {0}".format(ssxx))
            logger.debug("Sum of squared yy errors: {0}".format(ssyy))
            if ssxx*ssyy == 0: # Handle an irrational number
                return 0
            else:
                return ssxy/np.sqrt(ssxx*ssyy)



def euclidean(fname1, fname2):
    from math import sqrt
    if type(fname1) is not str:
        raise TypeError("kmerdb.distance.euclidean expects a str as its first positional argument")
    elif type(fname2) is not str:
        raise TypeError("kmerdb.distance.euclidean expects a str as its second positional argument")
    k = None
    logger.debug("Custom euclidean distance starting...")
    sum_of_squared_differences = 0
    with fileutil.open(fname1, mode='r') as kdb1:
        with fileutil.open(fname2, mode='r') as kdb2:
            if k is None:
                k = kdb1.meatadata['k']
            if k != kdb1.metadata['k']:
                raise Exception("File '{0}' reported k = {1} instead of k = {2}".format(f, kdb1.metadata['k'], k))
            elif k != kdb2.metdata['k']:
                raise Exception("File '{0}' reported k = {1} instead of k = {2}".format(f, kdb2.metadata['k'], k))
            N = 4 ** k

            for kmer_id in range(N):
                line1 = next(kdb1)
                line2 = next(kdb2)
                _, x = (int(_x) for _x in line1.rstrip().split("\t"))
                _, y = (int(_y) for _y in line2.rstrip().split("\t"))
                sum_of_squared_differences += (x - y)**2
                logger.debug("Sqrt | squared differences ((x-y)^2) ")
                logger.info("Calculating custom euclidean distance function")
                logger.debug("Self care...")
            return sqrt(sum_of_squared_differences)


def spearman(x, y):
    if type(x) is not np.ndarray:
        raise TypeError("kmerdb.distance.spearman expects a Numpy array as its first positional argument")
    elif type(y) is not np.ndarray:
        raise TypeError("kmerdb.distance.spearman expects a Numpy array as its second positional argument")
    from scipy.stats import spearmanr
    cor, pval = spearmanr(x, b=y)
    logger.debug("Spearman calculation from SciPy")
    logger.info("Smooth, buttery Spearman correlation coefficients.")
    return cor, pval

def EMD(x, y):
    if type(x) is not np.ndarray:
        raise TypeError("kmerdb.distance.EMD expects a Numpy array as its first positional argument")
    elif type(y) is not np.ndarray:
        raise TypeError("kmerdb.distance.EMD expects a Numpy array as its second positional argument")
    from scipy.stats import wasserstein_distance
    return wasserstein_distance(x, y)

        
def hamming(k, x, y):
    sum = 0
    for i in range(len(x)):
        if x[i] == y[i]:
            sum += 1
    return (1/4**k) * sum



# def d2s(x, y):
#     if type(x) is not np.ndarray:
#         raise TypeError("kmerdb.distance.d2s expects a Numpy array as its first positional argument")
#     elif type(y) is not np.ndarray:
#         raise TypeError("kmerdb.distance.d2s expects a Numpy array as its second positional argument")

    
#     from kmerdb import kmer
#     import math
    
#     N = len(x)
#     k = int(math.log(N, 4))
#     total_kmers_x = np.sum(x)
#     total_kmers_y = np.sum(y)
#     #mono_x = dict([c, np.round(mono_x[c]/total_kmers_x, 2) for c in mono_x])
#     #mono_y = dict([c, np.round(mono_y[c]/total_kmers_y, 2) for c in mono_y])
#     mono_x = dict([c, mono_x[c]/float(total_kmers_x) for c in mono_x])
#     mono_y = dict([c, mono_y[c]/float(total_kmers_y) for c in mono_y])


#     def _d2s(ex, ey, xi, yi):
#         xi_ = xi - (N-k+1)*ex
#         yi_ = yi - (N-k+1)*ey

#         return (xi_ * yi_)/np.sqrt(np.square(xi_) + np.square(yi_))
    
#     s = 0
#     for i in range(N):
#         seq = kmer.kmer_to_id(i)
#         Ex = functools.reduce(lambda a,b: a*mono_x[b], list(seq), 1)
#         Ey = functools.reduce(lambda a,b: a*mono_y[b], list(seq), 1)
#         s += _d2s(Ex, Ey, x[i], y[i])

#     return s



'''
   Copyright 2020 Matthew Ralston

   Licensed under the Apache License, Version 2.0 (the "License");
   you may not use this file except in compliance with the License.
   You may obtain a copy of the License at

       http://www.apache.org/licenses/LICENSE-2.0

   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   See the License for the specific language governing permissions and
   limitations under the License.

'''


import sys
import os
import numpy as np
import yaml
import math
import functools
import time


import logging
logger = logging.getLogger(__file__)



from kmerdb import fileutil, index, kmer
import Bio

def markov_probability(seq:Bio.SeqRecord.SeqRecord, kdbrdr:fileutil.KDBReader, kdbidx:index.IndexReader):
    """
    :param seq:
    :type SeqRecord: Bio.SeqRecord.SeqRecord
    :returns
    :rtype:
    """
    if not isinstance(seq, Bio.SeqRecord.SeqRecord):
        raise TypeError("kmerdb.probability.markov_probability expects a Bio.SeqRecord.SeqRecord as its first positional argument")

    elif not isinstance(kdbrdr, fileutil.KDBReader):
        raise TypeError("kmerdb.probability.markov_probability expects a kmerdb.fileutil.KDBReader as its second positionl argument")
    elif not isinstance(kdbidx, index.IndexReader):
        raise TypeError("kmerdb.probability.markov_probability expects a kmerdb.index.IndexReader as its third positional argument")

        
    k = kdbrdr.metadata['k']
    mononucleotides = kdbrdr.metadata['files'][0]['mononucleotides']
    N = 4**k
    total_kmers = sum([f['total_kmers'] for f in kdbrdr.metadata['files']])
    um = 1/N # This is the uniform frequency for the space
    if len(seq) < k:
        raise ValueError("kmerdb.probability.markov_probability expects the sequence to be at least k={0} in length".format(k))
    x = seq.seq[:k]
    kmer_id_x = kmer.kmer_to_id(x)
    if kmer_id_x > N:
        raise ValueError("kmerdb.probability.markov_probability expected the k-mer id {0} of the subsequence x='{1}' to have an id less than N = 4^k = {2}".format(kmer_id_x, x, N))
    if not isinstance(kdbidx.index, np.ndarray):
        raise ValueError("kmerdb.probability.markov_probability expects the kdbidx to be a Numpy array")
    elif kdbidx.index.size != N:
        raise ValueError("kmerdb.probability.markov_probability expects the kdbidx to have exactly {0} items, found {1}".format(N, kdbidx.size))

    
    kmer_id, count, neighbors = index.read_line(kdbrdr, kdbidx, kmer_id_x)
    if kmer_id is None or count is None or neighbors is None:
        logger.error("K-mer id: {0}\n".format(kmer_id_x, ))
        logger.error(kmer.id_to_kmer(kmer_id_x, k))
        time.sleep(1)
        raise RuntimeError("Index encountered an invalid initial k-mer id index value.")
    px =  count / float(total_kmers)
    #kmer_seq = kmer.id_to_kmer(km
    # Find the correct neighbor and calculate the transition probability
    prefix = x
    total = 0
    total2 = 0
    product = 1
    product2 = 1
    total_nucleotides = sum(mononucleotides.values())
    logger.debug(seq.seq)
    logger.debug(len(seq.seq))
    for i in range(k, len(seq.seq)):

        sum_qses = 0
        sum_aij = 0

        for char, idx in neighbors["suffixes"].items():
            kmer_id, count, _ = index.read_line(kdbrdr, kdbidx, idx)
            if kmer_id is None or count is None or _ is None:
                kmer_id = idx
                count = 0
                #raise ValueError("k-mer id '{0}' had an offset of zero in the index, it was not observed in the genome. Effective probability of sequence is 0 and Log-odds ratio of the sequence being generated from the k-mer profile is effectively 0 as well.".format(s))
            aij = mononucleotides[char] / total_nucleotides
            sum_aij += mononucleotides[char] / total_nucleotides
            sum_qses += count/float(total_kmers)
        # Prefix is the next prefix sequence to find a suffix for
        if i >= len(seq.seq):
            break
            #raise ValueError("kmerdb.probability.markov_probability encountered a value of i={0} larger than the length of the sequence {1}".format(i, len(seq.seq)))
        elif seq.seq[i] == "":
            logger.warning("Reached the end of the sequence, i={0} is producing no letters".format(i))
            break
        else:
            logger.debug("{0} : prefix: '{1}' + '{2}'".format(i, prefix[1:], seq.seq[i]))
            prefix = prefix[1:] + seq.seq[i]

        if len(prefix) != k:
            break
        else:
            logger.debug("New k-mer: '{0}'".format(prefix))
            logger.debug("Length of prefix: {0}".format(len(prefix)))
            try:
                kmerid = kmer.kmer_to_id(prefix)
            except KeyError as e:
                logger.debug(i)
                logger.debug("this should be the new letter: '{0}'".format(seq.seq[i]))
                logger.debug("this should be the existing prefix: '{0}'".format(prefix))
                raise e
            logger.debug("Iteration number {0}, calculating probabilities for k-mer id={1} seq={2}".format(i, kmerid, prefix))
            logger.debug("Reading counts/neighbors from file according to k-mer id")
            # We take the prefixes count and new neighbors
            kmer_id, count, neighbors = index.read_line(kdbrdr, kdbidx, kmerid)

            # Now we calculate qt similar to px, the frequency (count / total number of k-mers, read from the metadata)
        
            qt = count/float(total_kmers)
            # Now we calculate the transition probability of sequence s to sequence t
            # as qt divided by the sum of each qsc
            ast = qt / sum_qses

            # For the final log odds ratio, we sum the logs of the transition probabilities
            total += math.log10(ast)
            total2 += math.log10(sum_aij) # The ratio of the uniform frequency to the first order Markov background
            # For the final probability, we multiply the transition probabilities
            product *= ast
            product2 *= aij
    # Fianlly qxis represents our null model in a way.
    # Suppose that the null model consisted of the same number of k-mer as observed,
    # but uniformly distributed amongst all the k-mers
    qxis = [um for x in range(k, len(seq.seq))]
    # We put this on the bottom of the Log-odds Ratio because the null model contrasts the transition probabilities
    null_model = math.log10(um) + total2
    lor = (math.log10(px) + total) / null_model
    # 
    pseq = px*product#/functools.reduce(lambda x,y: x*y, qxis)
    prand = product2/N
    return {"seq": seq, "log_odds_ratio": lor, "p_of_seq": pseq}



'''
   Copyright 2020 Matthew Ralston

   Licensed under the Apache License, Version 2.0 (the "License");
   you may not use this file except in compliance with the License.
   You may obtain a copy of the License at

       http://www.apache.org/licenses/LICENSE-2.0

   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   See the License for the specific language governing permissions and
   limitations under the License.

'''


import sys
import os
import copy
import gzip
import yaml
import numpy as np
#import jsonschema
sys.path.append('..')

from kmerdb import fileutil


import logging
logger = logging.getLogger(__file__)



def open(filepath, mode="r", k=None, idx=None):
    if type(filepath) is not str:
        raise TypeError("kmerdb.index.open expects a str as its first positional argument")
    elif type(mode) is not str:
        raise TypeError("kmerdb.index.open expects a str as its second positional argument")
    elif "w" in mode and type(idx) is not np.ndarray:
        if "x" in mode and type(k) is not int: # Specifically, if its not a
            raise TypeError("kmerdb.index.open in create mode expects an int as its first *args argument")
            
    elif "w" in mode and type(k) is not int:
        raise TypeError("kmerdb.index.open in write mode expects an int as its second *args argument")
    modes = set(list(mode.lower()))
    if modes - set("xrwbt") or len(mode) > len(modes):
        raise ValueError("invalid mode: {}".format(mode))

    creating = "x" in modes
    reading  = "r" in modes
    writing  = "w" in modes
    binary   = "b" in modes
    text     = "t" in modes

    if text and binary:
        raise ValueError("can't have text and binary mode simultaneously")
    elif not (creating or reading or writing):
        raise ValueError("must have exactly one or read/write")

    if "r" in mode.lower():
        return IndexReader(filepath) # Will attempt to read this as an index file
    elif "x" in mode.lower() or set(list("xw")).intersection(modes):
        logger.info("Building index from file and constructing (mode='x') new .kdbi file")
        return IndexBuilder(filepath, k) # Creates a new .kdbi exactly named for the kdb file
    elif "w" in mode.lower():
        logger.info("Writing index to file instead of building index")
        write_index(idx, indexfile, k) # index, indexfile, k


class IndexReader:
    def __init__(self, indexfile:str):
        if type(indexfile) is not str:
            raise TypeError("kmerdb.index.IndexReader expects a str as its first positional argument")
        elif not os.path.exists(indexfile):
            raise TypeError("kmerdb.index.IndexReader expects an existing .kdbi index file as its first positional argument")
        elif not is_gz_file(indexfile):
            raise IOError("kmerdb.index.IndexReader expects a gzip compressed index file as its first positional argument")
        

        idx = np.array([], dtype="int64")
        #i = 0
        with gzip.open(indexfile, 'rt') as ifile:
            first_line = ifile.readline().rstrip()
            try:
                logger.debug("First line of index:\n{0}".format(first_line))
                k = int(first_line)
                idx = np.zeros(4**k, dtype="int64")
            except ValueError as e:
                logger.debug(e)
                raise IOError("Could not determine k by casting the first line of the index as an integer. Improperly formatted index")
            i = 0
            for line in ifile:
                # if i == 0:
                #     logger.warning("i: {0}, line:\n{1}".format(i, line))

                #     sys.exit(1)
                #     pass
                # else:
                kmer_id, line_offset = [int(i) for i in line.rstrip().split("\t")]
                idx[kmer_id] = line_offset
                i += 1
        self.index = idx

    def __enter__(self):
        return self

    def __exit__(self, type, value, tb):
        del self.index
        return

class IndexBuilder:
    def __init__(self, kdbfile:str, k:int):
        if type(kdbfile) is not str:
            raise TypeError("kmerdb.index.IndexBuilder expects a str as its first positional argument")
        elif not os.path.exists(kdbfile):
            logger.error("Could not find {0}".format(kdbfile))
            raise IOError("kmderdb.index.IndexBuilder expects an existing .kdb file as its first positional argument")
        elif os.path.splitext(kdbfile)[-1] != ".kdb":
            raise IOError("kmerdb.index.IndexBuilder was passed a non-kdb filetype as its first positional argument, should be .kdb")
        elif type(k) is not int:
            raise TypeError("kmerdb.index.IndexBuilder expects an int as its second positional argument")
        N = 4**k
        line_index = np.zeros(N, dtype="int64")
        #line_index = np.array([], dtype="int64")
        #line_index = set() # Can only use set transiently as it might sort
        indexfile = kdbfile + "i"
        
        with fileutil.open(kdbfile, mode='r') as kdbrdr:
            pos = kdbrdr.tell() # This could be the byte encoding, since we're reading it as plain text

            logger.info("Reading the .kdb file '{0}' in mode 'r' to generate the index...".format(kdbrdr._filepath))
            logger.info("Top position in the file after metadata parsing is said to be 'with this as kdbrdr:' {0}".format(pos))
            logger.info("KDBReader logs the position in mode 'r' as 'header_offset', after appending blocks: {0}".format(kdbrdr.metadata["header_offset"])) # This could be a compressed offset
            logger.info("Asked for the position one more time in mode 'r' in 'with this' and received: {0}".format(kdbrdr.tell()))
            line_index[0] = kdbrdr.tell()

            # SOmething should be checked here
            #kdbrdr.seek(kdbrdr.metadata["header_offset"])
            #kdbrdr._load_block()
            i = 1
            this_line = line_index[0]
            logger.debug("0th index, points to the first row? : {0}".format(line_index[0]))
            for line in kdbrdr:
                try:
                    next_line = kdbrdr.tell()
                    kmer_id, count = [int(i) for i in line.rstrip().split("\t")[:-1]]
                    line_index[kmer_id] = this_line
                    kdbrdr.seek(this_line)
                    new_line = kdbrdr.readline()
                    if new_line == line:
                        logger.debug("i: {0}, K-mer id: {1}, count: {2}, index line: {3}, offset: {4}".format(i, kmer_id, count, i, this_line))
                    elif next_line == 0 or this_line == 0:
                        logger.warning("i: 0, k-mer id: {1}, this_line: {2}, next_line{3}".format(i, kmer_id, this_line, next_line))
                        raise ValueError("Encountered a line where kdbrdr.tell() is producing 0")
                    else:
                        logger.warning("This line was not reproducibly generated...")
                        logger.warning("i: {0}, K-mer id: {1}, offset: {2}".format(i, kmer_id, this_line))
                        sys.exit(1)
                        
                    this_line = next_line
                    if kdbrdr.tell() == 0:
                        raise IOError("kmerdb.index.IndexBuilder expects the kdbrdr to increment with each line")
                    # elif i > 10:
                    #     logger.error("Position in file: {0}".format(pos))
                    #     sys.exit(1)
                except IndexError as e: # The final 
                    logger.warning("{0}:\t{1}".format(i, line.rstrip()))
                    raise e # STILL DONT KNOW WHATS HAPPENING HERE
                i += 1
        write_index(line_index, indexfile, k)

        
def has_index(kdbfile):
    if type(kdbfile) is not str:
        raise TypeError("kmerdb.index.has_index expects a str as its first positional argument")
    elif not os.path.exists(kdbfile):
        raise IOError("kmerdb.index.has_index received a filename that does not exist")
    elif os.path.splitext(kdbfile)[-1] == ".kdb" and not os.path.exists(kdbfile + "i"):
        return False
    elif os.path.splitext(kdbfile)[-1] == ".kdb" and os.path.exists(kdbfile + "i"):
        return True
    elif os.path.splitext(kdbfile)[-1] != ".kdb":
        raise ValueError("unexpected file extension, not a .kdb file.")
    else:
        raise RuntimeError("kmerdb.index.has_index encountered unexpected user inputs:\nkdbfile: '{0}', exists: {1}".format(kdbfile, os.path.exists(kdbfile)))


def read_line(kdbrdr:fileutil.KDBReader, kdbidx:IndexReader, kmer_id):
    if not isinstance(kdbrdr, fileutil.KDBReader):
        raise TypeError("kmerdb.index.read_line expects a kmerdb.fileutil.KDBReader as its first positional argument")
    elif not isinstance(kdbidx, IndexReader):
        raise TypeError("kmerdb.index.read_line expects a kmerdb.index.IndexReader as its second positional argument")
    elif type(kmer_id) is not int:
        raise TypeError("kmerdb.index.read_line expects an int as its third positional argument")
    if kdbidx.index[kmer_id] == 0:
        logger.warning("Found an index value without an offset, in fact, it's in the header")
        return None, None, None
        #raise ValueError("Invalid index value 0")
    else:
        kdbrdr.seek(kdbidx.index[kmer_id])
        kmer_id, count, metadata = kdbrdr.read_line()
        return kmer_id, count, metadata

        
def is_gz_file(filepath):
    try:
        gzip
        
        with gzip.open(filepath, 'rb') as f:
            f.readline()
        return True
    except OSError as e:
        return False

def write_index(index:np.array, indexfile:str, k:int):
    if not isinstance(index, np.ndarray):
        raise TypeError("kmerdb.index.write_index expects a str as its first positional argument")
    elif type(indexfile) is not str:
        raise TypeError("kmerdb.index.write_index expects a str as its second positional argument")
    elif os.path.exists(indexfile):
        logger.warning("kmerdb.index.write_index is overwriting an existing index '{0}'...".format(indexfile))
        mode = "wt"
    elif type(k) is not int:
        raise TypeError("kmerdb.index.write_index expects an int as its third positional argument")
    else:
        mode = "xt"
    with gzip.open(indexfile, mode) as ofile:
        ofile.write("{0}\n".format(k))
        for i, line_offset in enumerate(index):
            if int(line_offset) == 0:
                logger.debug("K-mer {0} was not observed, not writing to file".format(i))
                pass
            else:
                ofile.write("{0}\t{1}\n".format(i, line_offset))

